from paegan.transport.controllers import BaseModelController, Runner
from paegan.transport.forcers import BaseForcer
from paegan.logger import logger

import paegan.transport.export as ex


class IPythonClusterModelController(BaseModelController):

    def listen_for_results(self, output_h5_file, total_particles):
        logger.info("Waiting for %i particle results" % total_particles)

        particles = []
        retrieved = 0

        while retrieved < total_particles:
            try:
                # IPython parallel View
                # self.result is an AsyncMapResult
                from IPython.parallel import TimeoutError
                try:
                    new_particles = self.result.get(timeout=1)
                except TimeoutError:
                    pass    # this is fine, get incremental progress below
                else:
                    particles = new_particles

                # progress is absolute, not incremental
                retrieved = self.result.progress
            except StopIteration:
                assert retrieved >= total_particles
                break
            except:
                logger.exception("Particle has FAILED!!")
                continue

            # We multiply by 90 here to save 10% for the exporting
            logger.progress((round((float(retrieved) / total_particles) * 90., 1), "%s Particle(s) complete" % retrieved))

        results = ex.ResultsPyTable(output_h5_file)
        for p in particles:
            for x in range(len(p.locations)):
                results.write(p.timestep_index_dump(x))
        results.compute()
        results.close()

        return

    def start_tasks(self, **kwargs):

        try:
            logger.info('Adding %i particles as tasks' % self.total_particle_count())
            tasks = []

            for part in self.particles:
                forcer = BaseForcer(self.hydrodataset,
                                    particle=part,
                                    common_variables=self.common_variables,
                                    timevar=self.timevar,
                                    times=self.times,
                                    start_time=self.start,
                                    models=self._models,
                                    release_location_centroid=self.reference_location.point,
                                    usebathy=self._use_bathymetry,
                                    useshore=self._use_shoreline,
                                    usesurface=self._use_seasurface,
                                    reverse_distance=self.reverse_distance,
                                    bathy_path=self.bathy_path,
                                    shoreline_path=self.shoreline_path,
                                    shoreline_feature=self.shoreline_feature,
                                    time_method=self.time_method,
                                    shoreline_index_buffer=self.shoreline_index_buffer)
                tasks.append(forcer)

            logger.progress((5, 'Running model'))
            return self.pool.map_async(Runner(), tasks)

        except Exception:
            logger.exception("Something didn't start correctly!")
            raise



import os
import time as timer
import math
import time
import json
import logging

import textwrap

import numpy as np

from paegan.location4d import Location4D

from paegan.transport.utils.asatransport import AsaTransport
from paegan.transport.shoreline import Shoreline
from paegan.transport.bathymetry import Bathymetry

from paegan.cdm.dataset import CommonDataset
from paegan.cdm.timevar import date2num

from paegan.logger import logger


class BaseForcer(object):

    def __init__(self, hydrodataset, **kwargs):

        """
        part, common_variables, timevar, times, start_time, models,
        release_location_centroid, usebathy, useshore, usesurface,
        get_data, n_run, read_lock, has_read_lock, read_count,
        point_get, data_request_lock, has_data_request_lock, reverse_distance=None, bathy=None,
        shoreline_path=None, shoreline_feature=None, time_method=None, caching=None, redis_url=None, redis_results_channel=None, shoreline_index_buffer=None):

            This is the task/class/object/job that forces an
            individual particle and communicates with the
            other particles and data controller for local
            cache updates
        """
        assert hydrodataset is not None

        # Common parameters
        self.hydrodataset               = hydrodataset
        self.bathy_path                 = kwargs.get("bathy_path")
        self.release_location_centroid  = kwargs.get("release_location_centroid")
        self.particle                   = kwargs.get("particle")
        self.times                      = kwargs.get("times")
        self.timevar                    = kwargs.get("timevar", None)
        self.start_time                 = kwargs.get("start_time")
        self.models                     = kwargs.get("models", [])
        self.usebathy                   = kwargs.get("usebathy", False)
        self.useshore                   = kwargs.get("useshore", False)
        self.usesurface                 = kwargs.get("usesurface", True)
        self.shoreline_path             = kwargs.get("shoreline_path")
        self.shoreline_feature          = kwargs.get("shoreline_feature", None)
        self.shoreline_index_buffer     = kwargs.get("shoreline_index_buffer", 0.1)
        self.time_method                = kwargs.get("time_method", "nearest")
        self.reverse_distance           = kwargs.get("reverse_distance", 500)

        # Redis for results
        self.redis_url                  = kwargs.get("redis_url", None)
        self.redis_results_channel      = kwargs.get("redis_results_channel", None)

        # Set common variable names
        self.common_variables = kwargs.get("common_variables")
        self.uname      = self.common_variables.get("u", None)
        self.vname      = self.common_variables.get("v", None)
        self.wname      = self.common_variables.get("w", None)
        self.temp_name  = self.common_variables.get("temp", None)
        self.salt_name  = self.common_variables.get("salt", None)
        self.xname      = self.common_variables.get("x", None)
        self.yname      = self.common_variables.get("y", None)
        self.zname      = self.common_variables.get("z", None)
        self.tname      = self.common_variables.get("time", None)

        self.active     = None

    def load_initial_dataset(self):
        """
        Initialize self.dataset, then close it
        A cacher will have to wrap this in locks, while a straight runner will not.
        """
        try:
            self.dataset = CommonDataset.open(self.hydrodataset)
            if self.timevar is None:
                self.timevar = self.dataset.gettimevar(self.common_variables.get("u"))
        except Exception:
            logger.warn("No source dataset: %s.  Particle exiting" % self.hydrodataset)
            raise

    def boundary_interaction(self, **kwargs):
        """
            Returns a list of Location4D objects
        """
        particle = kwargs.pop('particle')
        starting = kwargs.pop('starting')
        ending   = kwargs.pop('ending')

        # shoreline
        if self.useshore:
            intersection_point = self._shoreline.intersect(start_point=starting.point, end_point=ending.point)
            if intersection_point is not None:
                # Set the intersection point.
                hitpoint = Location4D(point=intersection_point['point'], time=starting.time + (ending.time - starting.time))
                particle.location = hitpoint

                # This relies on the shoreline to put the particle in water and not on shore.
                resulting_point = self._shoreline.react(start_point=starting,
                                                        end_point=ending,
                                                        hit_point=hitpoint,
                                                        reverse_distance=self.reverse_distance,
                                                        feature=intersection_point['feature'],
                                                        distance=kwargs.get('distance'),
                                                        angle=kwargs.get('angle'),
                                                        azimuth=kwargs.get('azimuth'),
                                                        reverse_azimuth=kwargs.get('reverse_azimuth'))
                ending.latitude = resulting_point.latitude
                ending.longitude = resulting_point.longitude
                ending.depth = resulting_point.depth
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s - hit the shoreline at %s.  Setting location to %s." % (particle.logstring(), hitpoint.logstring(),  ending.logstring()))

        # bathymetry
        if self.usebathy:
            if not particle.settled:
                bintersect = self._bathymetry.intersect(start_point=starting, end_point=ending)
                if bintersect:
                    pt = self._bathymetry.react(type='reverse', start_point=starting, end_point=ending)
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("%s - hit the bottom at %s.  Setting location to %s." % (particle.logstring(), ending.logstring(), pt.logstring()))
                    ending.latitude = pt.latitude
                    ending.longitude = pt.longitude
                    ending.depth = pt.depth

        # sea-surface
        if self.usesurface:
            if ending.depth > 0:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s - rose out of the water.  Setting depth to 0." % particle.logstring())
                ending.depth = 0

        particle.location = ending

    def get_nearest_data(self, i):
        """ Note: self.dataset.opennc() must be called before calling this function.
            This is because the caching forcer must close it everytime, while a non caching
            forcer can leave the dataset open.
        """
        try:
            # Grab data at time index closest to particle location
            u = np.mean(np.mean(self.dataset.get_values('u', timeinds=[np.asarray([i])], point=self.particle.location )))
            v = np.mean(np.mean(self.dataset.get_values('v', timeinds=[np.asarray([i])], point=self.particle.location )))
            # if there is vertical velocity inthe dataset, get it
            if 'w' in self.dataset.nc.variables:
                w = np.mean(np.mean(self.dataset.get_values('w', timeindsf=[np.asarray([i])], point=self.particle.location )))
            else:
                w = 0.0
            # If there is salt and temp in the dataset, get it
            if self.temp_name is not None and self.salt_name is not None:
                temp = np.mean(np.mean(self.dataset.get_values('temp', timeinds=[np.asarray([i])], point=self.particle.location )))
                salt = np.mean(np.mean(self.dataset.get_values('salt', timeinds=[np.asarray([i])], point=self.particle.location )))

            # Check for nans that occur in the ocean (happens because
            # of model and coastline resolution mismatches)
            if np.isnan(u).any() or np.isnan(v).any() or np.isnan(w).any():
                # Take the mean of the closest 4 points
                # If this includes nan which it will, result is nan
                uarray1 = self.dataset.get_values('u', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                varray1 = self.dataset.get_values('v', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                if 'w' in self.dataset.nc.variables:
                    warray1 = self.dataset.get_values('w', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                    w = warray1.mean()
                else:
                    w = 0.0

                if self.temp_name is not None and self.salt_name is not None:
                    temparray1 = self.dataset.get_values('temp', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                    saltarray1 = self.dataset.get_values('salt', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                    temp = temparray1.mean()
                    salt = saltarray1.mean()
                u = uarray1.mean()
                v = varray1.mean()

            if self.temp_name is None:
                temp = np.nan
            if self.salt_name is None:
                salt = np.nan

        except Exception:
            logger.exception("Could not retrieve data.")
            raise

        return u, v, w, temp, salt

    def get_linterp_data(self, i, currenttime):
        """ Note: self.dataset.opennc() must be called before calling this function.
            This is because the caching forcer must close it everytime, while a non caching
            forcer can leave the dataset open.
        """
        try:
            # Grab data at time index closest to particle location
            u = [np.mean(np.mean(self.dataset.get_values('u', timeinds=[np.asarray([i])], point=self.particle.location ))),
                 np.mean(np.mean(self.dataset.get_values('u', timeinds=[np.asarray([i+1])], point=self.particle.location )))]
            v = [np.mean(np.mean(self.dataset.get_values('v', timeinds=[np.asarray([i])], point=self.particle.location ))),
                 np.mean(np.mean(self.dataset.get_values('v', timeinds=[np.asarray([i+1])], point=self.particle.location )))]
            # if there is vertical velocity inthe dataset, get it
            if 'w' in self.dataset.nc.variables:
                w = [np.mean(np.mean(self.dataset.get_values('w', timeinds=[np.asarray([i])], point=self.particle.location ))),
                     np.mean(np.mean(self.dataset.get_values('w', timeinds=[np.asarray([i+1])], point=self.particle.location )))]
            else:
                w = [0.0, 0.0]
            # If there is salt and temp in the dataset, get it
            if self.temp_name is not None and self.salt_name is not None:
                temp = [np.mean(np.mean(self.dataset.get_values('temp', timeinds=[np.asarray([i])], point=self.particle.location ))),
                        np.mean(np.mean(self.dataset.get_values('temp', timeinds=[np.asarray([i+1])], point=self.particle.location )))]
                salt = [np.mean(np.mean(self.dataset.get_values('salt', timeinds=[np.asarray([i])], point=self.particle.location ))),
                        np.mean(np.mean(self.dataset.get_values('salt', timeinds=[np.asarray([i+1])], point=self.particle.location )))]

            # Check for nans that occur in the ocean (happens because
            # of model and coastline resolution mismatches)
            if np.isnan(u).any() or np.isnan(v).any() or np.isnan(w).any():
                # Take the mean of the closest 4 points
                # If this includes nan which it will, result is nan
                uarray1 = self.dataset.get_values('u', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                varray1 = self.dataset.get_values('v', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                uarray2 = self.dataset.get_values('u', timeinds=[np.asarray([i+1])], point=self.particle.location, num=2)
                varray2 = self.dataset.get_values('v', timeinds=[np.asarray([i+1])], point=self.particle.location, num=2)
                if 'w' in self.dataset.nc.variables:
                    warray1 = self.dataset.get_values('w', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                    warray2 = self.dataset.get_values('w', timeinds=[np.asarray([i+1])], point=self.particle.location, num=2)
                    w = [warray1.mean(), warray2.mean()]
                else:
                    w = [0.0, 0.0]

                if self.temp_name is not None and self.salt_name is not None:
                    temparray1 = self.dataset.get_values('temp', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                    saltarray1 = self.dataset.get_values('salt', timeinds=[np.asarray([i])], point=self.particle.location, num=2)
                    temparray2 = self.dataset.get_values('temp', timeinds=[np.asarray([i+1])], point=self.particle.location, num=2)
                    saltarray2 = self.dataset.get_values('salt', timeinds=[np.asarray([i+1])], point=self.particle.location, num=2)
                    temp = [temparray1.mean(), temparray2.mean()]
                    salt = [saltarray1.mean(), saltarray2.mean()]
                u = [uarray1.mean(), uarray2.mean()]
                v = [varray1.mean(), varray2.mean()]

            # Linear interp of data between timesteps
            currenttime = date2num(currenttime)
            timevar = self.timevar.datenum
            u = self.linterp(timevar[i:i+2], u, currenttime)
            v = self.linterp(timevar[i:i+2], v, currenttime)
            w = self.linterp(timevar[i:i+2], w, currenttime)
            if self.temp_name is not None and self.salt_name is not None:
                temp = self.linterp(timevar[i:i+2], temp, currenttime)
                salt = self.linterp(timevar[i:i+2], salt, currenttime)

            if self.temp_name is None:
                temp = np.nan
            if self.salt_name is None:
                salt = np.nan

        except Exception:
            logger.exception("Could not retrieve data.")
            raise

        return u, v, w, temp, salt

    def linterp(self, setx, sety, x):
        """
            Linear interp of model data values between time steps
        """
        if math.isnan(sety[0]) or math.isnan(setx[0]):
            return np.nan
        return sety[0] + (x - setx[0]) * ( (sety[1]-sety[0]) / (setx[1]-setx[0]) )

    def run(self):

        self.load_initial_dataset()

        redis_connection = None
        if self.redis_url is not None and self.redis_results_channel is not None:
            import redis
            redis_connection = redis.from_url(self.redis_url)

        # Setup shoreline
        self._shoreline = None
        if self.useshore is True:
            self._shoreline = Shoreline(path=self.shoreline_path, feature_name=self.shoreline_feature, point=self.release_location_centroid, spatialbuffer=self.shoreline_index_buffer)
            # Make sure we are not starting on land.  Raises exception if we are.
            self._shoreline.intersect(start_point=self.release_location_centroid, end_point=self.release_location_centroid)

        # Setup Bathymetry
        if self.usebathy is True:
            try:
                self._bathymetry = Bathymetry(file=self.bathy_path)
            except Exception:
                logger.exception("Could not load Bathymetry file: %s, using no Bathymetry for this run!" % self.bathy_path)
                self.usebathy = False

        # Calculate datetime at every timestep
        modelTimestep, newtimes = AsaTransport.get_time_objects_from_model_timesteps(self.times, start=self.start_time)

        if self.time_method == 'interp':
            time_indexs = self.timevar.nearest_index(newtimes, select='before')
        elif self.time_method == 'nearest':
            time_indexs = self.timevar.nearest_index(newtimes)
        else:
            logger.warn("Method for computing u,v,w,temp,salt not supported!")
        try:
            assert len(newtimes) == len(time_indexs)
        except AssertionError:
            logger.exception("Time indexes are messed up. Need to have equal datetime and time indexes")
            raise

        # Keep track of how much time we spend in each area.
        tot_boundary_time = 0.
        tot_model_time    = {}
        tot_read_data     = 0.
        for m in self.models:
            tot_model_time[m.name] = 0.

        # Set the base conditions
        # If using Redis, send the results
        if redis_connection is not None:
            redis_connection.publish(self.redis_results_channel, json.dumps(self.particle.timestep_dump()))

        # loop over timesteps
        # We don't loop over the last time_index because
        # we need to query in the time_index and set the particle's
        # location as the 'newtime' object.
        for loop_i, i in enumerate(time_indexs[0:-1]):

            if self.active and self.active.value is False:
                raise ValueError("Particle exiting due to Failure.")

            newloc = None

            st = time.clock()
            # Get the variable data required by the models
            if self.time_method == 'nearest':
                u, v, w, temp, salt = self.get_nearest_data(i)
            elif self.time_method == 'interp':
                u, v, w, temp, salt = self.get_linterp_data(i, newtimes[loop_i])
            else:
                logger.warn("Method for computing u,v,w,temp,salt is unknown. Only 'nearest' and 'interp' are supported.")
            tot_read_data += (time.clock() - st)

            # Get the bathy value at the particles location
            if self.usebathy is True:
                bathymetry_value = self._bathymetry.get_depth(self.particle.location)
            else:
                bathymetry_value = -999999999999999

            # Age the particle by the modelTimestep (seconds)
            # 'Age' meaning the amount of time it has been forced.
            self.particle.age(seconds=modelTimestep[loop_i])

            # loop over models - sort these in the order you want them to run
            for model in self.models:
                st = time.clock()
                movement = model.move(self.particle, u, v, w, modelTimestep[loop_i], temperature=temp, salinity=salt, bathymetry_value=bathymetry_value)
                newloc = Location4D(latitude=movement['latitude'], longitude=movement['longitude'], depth=movement['depth'], time=newtimes[loop_i+1])
                tot_model_time[m.name] += (time.clock() - st)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s - moved %.3f meters (horizontally) and %.3f meters (vertically) by %s with data from %s" % (self.particle.logstring(), movement['distance'], movement['vertical_distance'], model.__class__.__name__, newtimes[loop_i].isoformat()))
                if newloc:
                    st = time.clock()
                    self.boundary_interaction(particle=self.particle, starting=self.particle.location, ending=newloc,
                                              distance=movement['distance'], angle=movement['angle'],
                                              azimuth=movement['azimuth'], reverse_azimuth=movement['reverse_azimuth'],
                                              vertical_distance=movement['vertical_distance'], vertical_angle=movement['vertical_angle'])
                    tot_boundary_time += (time.clock() - st)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s - was forced by %s and is now at %s" % (self.particle.logstring(), model.__class__.__name__, self.particle.location.logstring()))

            self.particle.note = self.particle.outputstring()
            # Each timestep, save the particles status and environmental variables.
            # This keep fields such as temp, salt, halted, settled, and dead matched up with the number of timesteps
            self.particle.save()

            # If using Redis, send the results
            if redis_connection is not None:
                redis_connection.publish(self.redis_results_channel, json.dumps(self.particle.timestep_dump()))

        self.dataset.closenc()

        # We won't pull data for the last entry in locations, but we need to populate it with fill data.
        self.particle.fill_gap()

        if self.usebathy is True:
            self._bathymetry.close()

        if self.useshore is True:
            self._shoreline.close()

        logger.info(textwrap.dedent('''Particle %i Stats:
                          Data read: %f seconds
                          Model forcing: %s seconds
                          Boundary intersection: %f seconds''' % (self.particle.uid, tot_read_data, { s: '{:g} seconds'.format(f) for s, f in list(tot_model_time.items()) }, tot_boundary_time)))

        return self.particle

    def __call__(self, active):
        self.active = active
        return self.run()


class CachingForcer(BaseForcer):

    def __init__(self, *args, **kwargs):
        super(CachingForcer, self).__init__(*args, **kwargs)
        self.get_data               = kwargs.get("get_data")
        self.read_lock              = kwargs.get("read_lock")
        self.has_read_lock          = kwargs.get("has_read_lock")
        self.read_count             = kwargs.get("read_count")
        self.point_get              = kwargs.get("point_get")
        self.data_request_lock      = kwargs.get("data_request_lock")
        self.has_data_request_lock  = kwargs.get("has_data_request_lock")

    def load_initial_dataset(self):
        """
        Initialize self.dataset, then close it
        A cacher will have to wrap this in locks, while a straight runner will not.
        """
        try:
            with self.read_lock:
                self.read_count.value += 1
                self.has_read_lock.append(os.getpid())
            self.dataset = CommonDataset.open(self.hydrodataset)
            self.dataset.closenc()
        except Exception:
            logger.warn("No source dataset: %s.  Particle exiting" % self.hydrodataset)
            raise
        finally:
            with self.read_lock:
                self.read_count.value -= 1
                self.has_read_lock.remove(os.getpid())

    def __call__(self, active):

        if active.value is True:
            while self.get_data.value is True:
                logger.info("Waiting for DataController to start...")
                timer.sleep(5)
                pass

        return super(CachingForcer, self).__call__(active)

    def get_nearest_data(self, i):
        self.fill_cache_with_nearest_data(i)

        # Now that the cache is filled, get the actual data and return
        with self.read_lock:
            self.read_count.value += 1
            self.has_read_lock.append(os.getpid())
        try:
            self.dataset.opennc()
            return super(CachingForcer, self).get_nearest_data(i)
        except Exception:
            logger.exception("Could not retrieve data")
        finally:
            self.dataset.closenc()
            with self.read_lock:
                self.read_count.value -= 1
                self.has_read_lock.remove(os.getpid())

    def get_linterp_data(self, i, currenttime):
        self.fill_cache_with_linterp_data(i, currenttime)

        # Now that the cache is filled, get the actual data and return
        with self.read_lock:
            self.read_count.value += 1
            self.has_read_lock.append(os.getpid())
        try:
            self.dataset.opennc()
            return super(CachingForcer, self).get_linterp_data(i, currenttime)
        except Exception:
            logger.exception("Could not retrieve data")
        finally:
            self.dataset.closenc()
            with self.read_lock:
                self.read_count.value -= 1
                self.has_read_lock.remove(os.getpid())

    def need_data(self, i):
        """
            Method to test if cache contains the data that
            the particle needs
        """

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Checking cache for data availability at %s." % self.particle.location.logstring())

        try:
            # Tell the DataController that we are going to be reading from the file
            with self.read_lock:
                self.read_count.value += 1
                self.has_read_lock.append(os.getpid())

            self.dataset.opennc()
            # Test if the cache has the data we need
            # If the point we request contains fill values,
            # we need data
            cached_lookup = self.dataset.get_values('domain', timeinds=[np.asarray([i])], point=self.particle.location)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Type of result: %s" % type(cached_lookup))
                logger.debug("Double mean of result: %s" % np.mean(np.mean(cached_lookup)))
                logger.debug("Type of Double mean of result: %s" % type(np.mean(np.mean(cached_lookup))))
            if type(np.mean(np.mean(cached_lookup))) == np.ma.core.MaskedConstant:
                need = True
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("I NEED data.  Got back: %s" % cached_lookup)
            else:
                need = False
                logger.debug("I DO NOT NEED data")
        except Exception:
            # If the time index doesnt even exist, we need
            need = True
            logger.debug("I NEED data (no time index exists in cache)")
        finally:
            self.dataset.closenc()
            with self.read_lock:
                self.read_count.value -= 1
                self.has_read_lock.remove(os.getpid())

        return need  # Returns True if we need data or False if we dont

    def fill_cache_with_nearest_data(self, i):
        """
            Method to streamline request for data from cache,
            Uses nearest time to get u,v,w,temp,salt
        """
        if self.active.value is True:
            while self.get_data.value is True:
                logger.debug("Waiting for DataController to release cache file so I can read from it...")
                timer.sleep(2)
                pass

        if self.need_data(i):
            # Acquire lock for asking for data
            self.data_request_lock.acquire()
            self.has_data_request_lock.value = os.getpid()
            try:
                if self.need_data(i):

                    with self.read_lock:
                        self.read_count.value += 1
                        self.has_read_lock.append(os.getpid())

                    # Open netcdf file on disk from commondataset
                    self.dataset.opennc()
                    # Get the indices for the current particle location
                    indices = self.dataset.get_indices('u', timeinds=[np.asarray([i-1])], point=self.particle.location )
                    self.dataset.closenc()

                    with self.read_lock:
                        self.read_count.value -= 1
                        self.has_read_lock.remove(os.getpid())

                    # Override the time
                    self.point_get.value = [indices[0]+1, indices[-2], indices[-1]]
                    # Request that the data controller update the cache
                    # DATA CONTOLLER STARTS
                    self.get_data.value = True
                    # Wait until the data controller is done
                    if self.active.value is True:
                        while self.get_data.value is True:
                            logger.debug("Waiting for DataController to update cache...")
                            timer.sleep(2)
                            pass
            except Exception:
                raise
            finally:
                self.has_data_request_lock.value = -1
                self.data_request_lock.release()

    def fill_cache_with_linterp_data(self, i, currenttime):
        """
            Method to streamline request for data from cache,
            Uses linear interpolation bewtween timesteps to
            get u,v,w,temp,salt
        """
        if self.active.value is True:
            while self.get_data.value is True:
                logger.debug("Waiting for DataController to release cache file so I can read from it...")
                timer.sleep(2)
                pass

        if self.need_data(i+1):
            # Acquire lock for asking for data
            self.data_request_lock.acquire()
            self.has_data_request_lock.value = os.getpid()
            try:
                # Do I still need data?
                if self.need_data(i+1):

                    # Tell the DataController that we are going to be reading from the file
                    with self.read_lock:
                        self.read_count.value += 1
                        self.has_read_lock.append(os.getpid())

                    # Open netcdf file on disk from commondataset
                    self.dataset.opennc()
                    # Get the indices for the current particle location
                    indices = self.dataset.get_indices('u', timeinds=[np.asarray([i-1])], point=self.particle.location )
                    self.dataset.closenc()

                    with self.read_lock:
                        self.read_count.value -= 1
                        self.has_read_lock.remove(os.getpid())

                    # Override the time
                    # get the current time index data
                    self.point_get.value = [indices[0] + 1, indices[-2], indices[-1]]
                    # Request that the data controller update the cache
                    self.get_data.value = True
                    # Wait until the data controller is done
                    if self.active.value is True:
                        while self.get_data.value is True:
                            logger.debug("Waiting for DataController to update cache with the CURRENT time index")
                            timer.sleep(2)
                            pass

                    # Do we still need to get the next timestep?
                    if self.need_data(i+1):
                        # get the next time index data
                        self.point_get.value = [indices[0] + 2, indices[-2], indices[-1]]
                        # Request that the data controller update the cache
                        self.get_data.value = True
                        # Wait until the data controller is done
                        if self.active.value is True:
                            while self.get_data.value is True:
                                logger.debug("Waiting for DataController to update cache with the NEXT time index")
                                timer.sleep(2)
                                pass
            except Exception:
                logger.warn("Particle failed to request data correctly")
                raise
            finally:
                # Release lock for asking for data
                self.has_data_request_lock.value = -1
                self.data_request_lock.release()



import os
from paegan.cdm.dataset import CommonDataset
from paegan.location4d import Location4D

import numpy as np


class Bathymetry():
    def __init__(self, **kwargs):
        """
            Optional named arguments:
            * file (local path or dap to bathymetry netcdf file)

        """

        if kwargs.get("file", None) is not None:
            self._file = os.path.normpath(kwargs.pop('file'))
        else:
            raise ValueError("Must provide a path to the Bathymetry file")

        self._type = kwargs.pop("type", "hover")
        self._nc = CommonDataset.open(self._file)
        self._bathy_name = kwargs.pop("bathy", "z")

    def close(self):
        self._nc.closenc()

    def intersect(self, **kwargs):
        """
            Intersect Point and Bathymetry
            returns bool
        """
        end_point = kwargs.pop('end_point')
        depth = self.get_depth(location=end_point)
        # Bathymetry and a particle's depth are both negative down
        if depth < 0 and depth > end_point.depth:
            inter = True
        else:
            inter = False
        return inter

    def get_depth(self, location):
        return np.mean(np.mean(self._nc.get_values(self._bathy_name, point=location)))

    def react(self, **kwargs):
        """
            The time of recation is ignored hereTime is ignored here
            and should be handled by whatever called this function.
        """
        react_type = kwargs.get("type", self._type)

        if react_type == 'hover':
            return self.__hover(**kwargs)
        elif react_type == 'stick':
            pass
        elif react_type == 'reverse':
            return self.__reverse(**kwargs)
        else:
            raise ValueError("Bathymetry interaction type not supported")

    def __hover(self, **kwargs):
        """
            This hovers the particle 1m above the bathymetry WHERE IT WOULD HAVE ENDED UP.
            This is WRONG and we need to compute the location that it actually hit
            the bathymetry and hover 1m above THAT.
        """
        end_point = kwargs.pop('end_point')
        # The location argument here should be the point that intersected the bathymetry,
        # not the end_point that is "through" the bathymetry.
        depth = self.get_depth(location=end_point)
        return Location4D(latitude=end_point.latitude, longitude=end_point.longitude, depth=(depth + 1.))

    def __reverse(self, **kwargs):
        """
            If we hit the bathymetry, set the location to where we came from.
        """
        start_point = kwargs.pop('start_point')
        return Location4D(latitude=start_point.latitude, longitude=start_point.longitude, depth=start_point.depth)



import json
import math
from paegan.transport.models.behaviors.diel import Diel
from paegan.transport.models.behaviors.taxis import Taxis
from paegan.transport.models.behaviors.capability import Capability
from paegan.transport.models.behaviors.settlement import Settlement
from paegan.transport.models.base_model import BaseModel
from paegan.location4d import Location4D
from paegan.transport.utils.asatransport import AsaTransport
import operator
from datetime import timedelta

from paegan.logger import logger


class LifeStage(BaseModel):

    def __init__(self, **kwargs):

        if 'json' in kwargs or 'data' in kwargs:
            data = {}
            try:
                data = json.loads(kwargs['json'])
            except Exception:
                try:
                    data = kwargs.get('data')
                except Exception:
                    pass

            self.name = data.get('name',None)
            self.linear_a = data.get('linear_a', None)
            self.linear_b = data.get('linear_b', None)
            # duration is in days
            self.duration = data.get('duration', None)
            self.diel = [Diel(data=d) for d in data.get('diel')]
            self.taxis = [Taxis(data=t) for t in data.get('taxis')]
            self.capability = None
            if data.get('capability', None) is not None:
                self.capability = Capability(data=data.get('capability'))
            self.settlement = None
            if data.get('settlement', None) is not None:
                self.settlement = Settlement(data=data.get('settlement'))

    def get_active_diel(self, loc4d):
        active_diel = None
        if len(self.diel) > 0:
            particle_time = loc4d.time
            # Find the closests Diel that the current particle time is AFTER, and set it to the active_diel
            closest = None
            closest_seconds = None
            for ad in self.diel:
                # To handle thecase where a particle at 1:00am only looks at Diels for that
                # day and does not act upon Diel from the previous day at, say 11pm, check
                # both today's Diel times and the particles current days Diel times.
                yesterday = Location4D(location=loc4d)
                yesterday.time = yesterday.time - timedelta(days=1)

                times = [ad.get_time(loc4d=loc4d), ad.get_time(loc4d=yesterday)]
                for t in times:
                    if t <= particle_time:
                        seconds = (particle_time - t).total_seconds()
                        if closest is None or seconds < closest_seconds:
                            closest = ad
                            closest_seconds = seconds

                del yesterday

            active_diel = closest

        return active_diel

    def move(self, particle, u, v, w, modelTimestep, **kwargs):

        temp = kwargs.get('temperature', None)
        salt = kwargs.get('salinity', None)

        logger.debug("Temp: %.4f, Salt: %.4f" % (temp, salt))

        # IMPORTANT:
        # If we got no data from the model, we are using the last available value stored in the particles!
        if (temp is None) or (temp is not None and math.isnan(temp)):
            temp = particle.last_temp()
        if (salt is None) or (salt is not None and math.isnan(salt)):
            salt = particle.last_salt()

        particle.temp = temp
        particle.salt = salt

        # Grow the particle.  Growth affects which lifestage the particle is in.
        growth = 0.
        do_duration_growth = True
        modelTimestepDays = modelTimestep / 60. / 60. / 24.
        if self.linear_a is not None and self.linear_b is not None:
            if particle.temp is not None and not math.isnan(particle.temp):
                # linear growth, compute q = t / (Ax+B)
                # Where timestep t (days), at temperature x (deg C), proportion of stage completed (q)
                growth = modelTimestepDays / (self.linear_a * particle.temp + self.linear_b)
                particle.grow(growth)
                do_duration_growth = False
            else:
                logger.debug("No temperature found for Particle %s at this location and timestep, skipping linear temperature growth and using duration growth" % particle.uid)
                pass

        if do_duration_growth is True:
            growth = modelTimestepDays / self.duration
            particle.grow(growth)

        active_diel = self.get_active_diel(particle.location)

        # Run the active diel behavior and all of the taxis behaviors
        # u, v, and w store the continuous results from all of the behavior models.
        u = 0
        v = 0
        w = 0

        behaviors_to_run = [_f for _f in [self.settlement] + [active_diel] + self.taxis if _f]
        # Sort these in the order you want them to be run.

        try:
            vss = self.capability.calculated_vss
        except AttributeError:
            logger.debug("No VSS found, vertical behaviors will not act upon particle")
            vss = 0

        for behave in behaviors_to_run:
            behave_results = behave.move(particle, 0, 0, vss, modelTimestep, **kwargs)
            u += behave_results['u']
            v += behave_results['v']
            w += behave_results['w']

        # Do the calculation to determine the new location after running the behaviors
        result = AsaTransport.distance_from_location_using_u_v_w(u=u, v=v, w=w, timestep=modelTimestep, location=particle.location)
        result['u'] = u
        result['v'] = v
        result['w'] = w
        return result


class DeadLifeStage(LifeStage):
    def __init__(self, **kwargs):
        super(DeadLifeStage,self).__init__(**kwargs)

    def move(self, particle, u, v, w, modelTimestep, **kwargs):
        """ I'm dead, so no behaviors should act on me """

        # Kill the particle if it isn't settled and isn't already dead.
        if not particle.settled and not particle.dead:
            particle.die()

        # Still save the temperature and salinity for the model output
        temp = kwargs.get('temperature', None)
        if temp is not None and math.isnan(temp):
            temp = None
        particle.temp = temp

        salt = kwargs.get('salinity', None)
        if salt is not None and math.isnan(salt):
            salt = None
        particle.salt = salt

        u = 0
        v = 0
        w = 0

        # Do the calculation to determine the new location
        result = AsaTransport.distance_from_location_using_u_v_w(u=u, v=v, w=w, timestep=modelTimestep, location=particle.location)
        result['u'] = u
        result['v'] = v
        result['w'] = w
        return result



from paegan.transport.utils.asasuncycles import SunCycles
from paegan.location4d import Location4D
from paegan.utils.asamath import AsaMath
from datetime import datetime, timedelta
import pytz
import json
from paegan.transport.models.base_model import BaseModel

from paegan.logger import logger


class Diel(BaseModel):

    PATTERN_CYCLE        = "cycles"
    PATTERN_SPECIFICTIME = "specifictime"

    CYCLE_SUNRISE = "sunrise"
    CYCLE_SUNSET  = "sunset"

    HOURS_PLUS  = "+"
    HOURS_MINUS = "-"

    def __init__(self, **kwargs):

        if 'json' in kwargs or 'data' in kwargs:
            data = {}
            try:
                data = json.loads(kwargs['json'])
            except Exception:
                try:
                    data = kwargs.get('data')
                except Exception:
                    pass

            self.pattern = data.get('type',None)
            self.cycle = data.get('cycle', None)
            t = data.get('time', None)
            if t is not None:
                # time is in microseconds in JSON
                t = datetime.utcfromtimestamp(t / 1000)
            self.time = t
            self.plus_or_minus = data.get('plus_or_minus', None)
            # Convert from positive down to negative down
            self.min_depth = data.get('min') * -1.
            self.max_depth = data.get('max') * -1.
            self.time_delta = data.get('hours', None)

    def get_pattern(self):
        return self._pattern
    def set_pattern(self, c):
        self._pattern = c
    pattern = property(get_pattern, set_pattern)

    def get_time_delta(self):
        return self._time_delta
    def set_time_delta(self, hours):
        if AsaMath.is_number(hours):
            hours = int(hours)
        self._time_delta = hours
    time_delta = property(get_time_delta, set_time_delta)

    def get_cycle(self):
        return self._cycle
    def set_cycle(self, cycle):
        self._cycle = cycle
    cycle = property(get_cycle, set_cycle)

    def get_min_depth(self):
        return self._min_depth
    def set_min_depth(self, min_depth):
        self._min_depth = float(min_depth)
    min_depth = property(get_min_depth, set_min_depth)

    def get_max_depth(self):
        return self._max_depth
    def set_max_depth(self, max_depth):
        self._max_depth = float(max_depth)
    max_depth = property(get_max_depth, set_max_depth)

    def get_plus_or_minus(self):
        return self._plus_or_minus
    def set_plus_or_minus(self, pom):
        if pom is not None:
            if pom != self.HOURS_PLUS and pom != self.HOURS_MINUS:
                raise ValueError("plus_or_minus must equal '%s' or '%s'" % (self.HOURS_PLUS, self.HOURS_MINUS))
        self._plus_or_minus = pom
    plus_or_minus = property(get_plus_or_minus, set_plus_or_minus)

    def set_time(self, t):
        if t is not None:
            if not isinstance(t, datetime):
                raise ValueError("Time value must be a DateTime")
            t = t.replace(tzinfo=pytz.utc)
        self._time = t
    def get_time(self, loc4d=None):
        """
            Based on a Location4D object and this Diel object, calculate
            the time at which this Diel migration is actually happening
        """
        if loc4d is None:
            raise ValueError("Location4D object can not be None")

        if self.pattern == self.PATTERN_CYCLE:
            c = SunCycles.cycles(loc=loc4d)
            if self.cycle == self.CYCLE_SUNRISE:
                r = c[SunCycles.RISING]
            elif self.cycle == self.CYCLE_SUNSET:
                r = c[SunCycles.SETTING]
            td = timedelta(hours=self.time_delta)
            if self.plus_or_minus == self.HOURS_PLUS:
                r = r + td
            elif self.plus_or_minus == self.HOURS_MINUS:
                r = r - td
            return r
        elif self.pattern == self.PATTERN_SPECIFICTIME:
            return self._time.replace(year=loc4d.time.year, month=loc4d.time.month, day=loc4d.time.day)
    time = property(get_time, set_time)

    def move(self, particle, u, v, w, modelTimestep, **kwargs):

        # If the particle is settled, don't move it anywhere
        if particle.settled:
            return { 'u': 0, 'v': 0, 'w': 0 }

        # If the particle is halted (but not settled), don't move it anywhere
        if particle.halted:
            return { 'u': 0, 'v': 0, 'w': 0 }

        # How far could I move?  We don't want to overshoot our desired depth.
        vertical_potential = w * modelTimestep

        """
            This only works if min is less than max.
            No checks are done here, so it should be done before
            calling this function.
        """

        """ I'm below my desired max depth, so i need to go down

            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            -------------------------------------- min
            -------------------------------------- max
                                x                  me
            ______________________________________
        """
        if particle.location.depth < self.max_depth:
            logger.debug("DIEL: %s - Moving UP to desired depth from %f" % (self.logstring(), particle.location.depth))

            # If we are going to overshoot the desired minimum depth,
            # calculate a new w to land in the middle of the range.
            overshoot_distance = abs(particle.location.depth - self.min_depth)
            if overshoot_distance < abs(vertical_potential):
                halfway_distance = abs((self.max_depth - self.min_depth) / 2)
                w = ((overshoot_distance - halfway_distance) / modelTimestep)

            return { 'u': u, 'v': v, 'w': w }

        """ I'm above my desired min depth, so i need to go down

            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                                x                  me
            -------------------------------------- min
            -------------------------------------- max
            ______________________________________
        """
        if particle.location.depth > self.min_depth:
            logger.debug("DIEL: %s - Moving DOWN to desired depth from %f" % (self.logstring(), particle.location.depth))

            # If we are going to overshoot the desired maximum depth,
            # calculate a new w to land in the middle of the range.
            overshoot_distance = abs(particle.location.depth - self.max_depth)
            if overshoot_distance < abs(vertical_potential):
                halfway_distance = abs((self.max_depth - self.min_depth) / 2)
                w = ((overshoot_distance - halfway_distance) / modelTimestep)

            return { 'u': u, 'v': v, 'w': -w }

        """ I'm in my desired depth, so I'm just gonna chill here

            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            -------------------------------------- min
                                x                  me
            -------------------------------------- max
            ______________________________________
        """
        return { 'u': u, 'v': v, 'w': 0 }

    def logstring(self):
        if self.pattern == self.PATTERN_CYCLE:
            return "At %s (%s%d hours) go to between %fm and %fm" % (self.cycle.upper(), self.plus_or_minus, self.time_delta, self.min_depth, self.max_depth)
        else:
            return "At %s (UTC) go to between %fm and %fm" % (self._time.strftime("%H:%M"), self.min_depth, self.max_depth)



import json

from paegan.logger import logger

from paegan.location4d import Location4D

class Settlement(object):

    def __init__(self, **kwargs):

        if 'json' in kwargs or 'data' in kwargs:
            data = {}
            try:
                data = json.loads(kwargs['json'])
            except Exception:
                try:
                    data = kwargs.get('data')
                except Exception:
                    pass

            try:
                # Convert positive down to negative down
                self.upper = data.get('upper') * -1.
                self.lower = data.get('lower') * -1.
                self.type = data.get('type')
            except Exception:
                raise ValueError("A settlement must consist of a 'type' and 'upper / 'lower' bounds.")

    def attempt(self, particle, depth):

        # We may want to have settlement affect the u/v/w in the future
        u = 0
        v = 0
        w = 0

        # If the particle is settled, don't move it anywhere
        if particle.settled:
            return (0,0,0)

        # A particle is negative down from the sea surface, so "-3" is 3 meters below the surface.
        # We are assuming here that the bathymetry is also negative down.

        if self.type.lower() == "benthic":
            # Is the sea floor within the upper and lower bounds?
            if self.upper >= depth >= self.lower:
                # Move the particle to the sea floor.
                # TODO: Should the particle just swim downwards?
                newloc = Location4D(location=particle.location)
                newloc.depth = depth
                particle.location = newloc
                particle.settle()
                logger.info("Particle %d settled in %s mode" % (particle.uid, self.type))
        elif self.type.lower() == "pelagic":
            # Are we are in enough water to settle
            # Ignore this bathymetry test since we would need a high resolution
            # dataset for this to work.
            #if self.upper >= depth:

            # Is the particle within the range?
            if self.upper >= particle.location.depth >= self.lower:
                # Just settle the particle
                particle.settle()
                logger.info("Particle %d settled in %s mode" % (particle.uid, self.type))
            else:
                logger.debug("Particle did NOT settle.  Depth conditions not met.  Upper limit: %d - Lower limit: %d - Particle: %d" % (self.upper, self.lower, particle.location.depth))

            #else:
            #    logger.info("Particle did NOT settle.  Water not deep enough.  Upper limit: %d - Bathymetry: %d" % (self.upper, depth))
        else:
            logger.warn("Settlement type %s not recognized, not trying to settle Particle %d." % (self.type, particle.uid))

        return (u,v,w)

    def __str__(self):
        return \
        """*** Settlement  ***
        Upper: %d
        Lower: %d
        Type: %s
        """ % (self.upper, self.lower, self.type)

    def move(self, particle, u, v, w, modelTimestep, **kwargs):

        bathymetry_value = kwargs.pop("bathymetry_value", None)
        if bathymetry_value is None:
            logger.debug("No bathymetry so can not attempt to settle particle")
            return { 'u': 0, 'v': 0, 'w': 0 }

        u,v,w = self.attempt(particle, bathymetry_value)

        return { 'u': u, 'v': v, 'w': w }



import logging

try:
    PROGRESS=15
    logging.PROGRESS = PROGRESS
    logging.addLevelName(PROGRESS, 'PROGRESS')
    def progress(self, message, *args, **kws):
        if self.isEnabledFor(PROGRESS):
            self._log(PROGRESS, message, args, **kws)
    logging.Logger.progress = progress
except:
    pass

__version__ = '1.0.5'



try:
    from __builtin__ import unicode as str
except ImportError:
    pass
import os
import sys
import time as timer
import math
import traceback
try:
    import Queue as queue
except ImportError:
    import queue
import logging
import multiprocessing

import numpy as np
import netCDF4

from paegan.location4d import Location4D

from paegan.transport.utils.asatransport import AsaTransport
from paegan.transport.shoreline import Shoreline
from paegan.transport.bathymetry import Bathymetry
from paegan.transport.exceptions import CachingDataControllerError

from paegan.cdm.dataset import CommonDataset

from paegan.transport.forcers import BaseForcer

from paegan.logger import logger


class Consumer(multiprocessing.Process):
    def __init__(self, task_queue, result_queue, n_run, nproc_lock, active=True, get_data=None, **kwargs):
        """
            This is the process class that does all the handling of queued tasks
        """
        multiprocessing.Process.__init__(self, **kwargs)
        self.task_queue     = task_queue
        self.result_queue   = result_queue
        self.n_run          = n_run
        self.nproc_lock     = nproc_lock
        self.active         = active
        self.get_data       = get_data

    def run(self):

        while True:

            try:
                next_task = self.task_queue.get(True, 10)
            except queue.Empty:
                logger.info("No tasks left to complete, closing %s" % self.name)
                break
            else:
                answer = (None, None)
                try:
                    answer = (1, next_task(self.active))
                except Exception:
                    logger.exception("Disabling Error")
                    if isinstance(next_task, CachingDataController):
                        answer = (-2, "CachingDataController")
                        # Tell the particles that the CachingDataController is releasing file
                        self.get_data.value = False
                        # The data controller has died, so don't process any more tasks
                        self.active.value = False
                    elif isinstance(next_task, BaseForcer):
                        answer = (-1, next_task.particle)
                    else:
                        logger.warn("Strange task raised an exception: %s" % str(next_task.__class__))
                        answer = (None, None)
                finally:
                    self.result_queue.put(answer)

                    self.nproc_lock.acquire()
                    self.n_run.value = self.n_run.value - 1
                    self.nproc_lock.release()

                    self.task_queue.task_done()


class CachingDataController(object):
    def __init__(self, hydrodataset, common_variables, n_run, get_data, write_lock, has_write_lock, read_lock, read_count,
                 time_chunk, horiz_chunk, times, start_time, point_get, start, **kwargs):
        """
            The data controller controls the updating of the
            local netcdf data cache
        """
        assert "cache_path" in kwargs
        self.cache_path = kwargs["cache_path"]
        self.caching = kwargs.get("caching", True)
        self.hydrodataset = hydrodataset
        if self.cache_path == self.hydrodataset and self.caching is True:
            raise CachingDataControllerError("Caching is set to True but the cache path and data path are the same.  Refusing to overwrite the data path.")

        self.n_run = n_run
        self.get_data = get_data
        self.write_lock = write_lock
        self.has_write_lock = has_write_lock
        self.read_lock = read_lock
        self.read_count = read_count
        self.inds = None  # np.arange(init_size+1)
        self.time_size = time_chunk
        self.horiz_size = horiz_chunk
        self.point_get = point_get
        self.start_time = start_time
        self.times = times
        self.start = start

        # Set common variable names
        self.uname = common_variables.get("u", None)
        self.vname = common_variables.get("v", None)
        self.wname = common_variables.get("w", None)
        self.temp_name = common_variables.get("temp", None)
        self.salt_name = common_variables.get("salt", None)
        self.xname = common_variables.get("x", None)
        self.yname = common_variables.get("y", None)
        self.zname = common_variables.get("z", None)
        self.tname = common_variables.get("time", None)

    def get_remote_data(self, localvars, remotevars, inds, shape):
        """
            Method that does the updating of local netcdf cache
            with remote data
        """
        # If user specifies 'all' then entire xy domain is
        # grabbed, default is 4, specified in the model controller
        if self.horiz_size == 'all':
            y, y_1 = 0, shape[-2]
            x, x_1 = 0, shape[-1]
        else:
            r = self.horiz_size
            x, x_1 = self.point_get.value[2]-r, self.point_get.value[2]+r+1
            y, y_1 = self.point_get.value[1]-r, self.point_get.value[1]+r+1
            x, x_1 = x[0], x_1[0]
            y, y_1 = y[0], y_1[0]
            if y < 0:
                y = 0
            if x < 0:
                x = 0
            if y_1 > shape[-2]:
                y_1 = shape[-2]
            if x_1 > shape[-1]:
                x_1 = shape[-1]

        # Update domain variable for where we will add data
        domain = self.local.variables['domain']

        if len(shape) == 4:
            domain[inds[0]:inds[-1]+1, 0:shape[1], y:y_1, x:x_1] = np.ones((inds[-1]+1-inds[0], shape[1], y_1-y, x_1-x))
        elif len(shape) == 3:
            domain[inds[0]:inds[-1]+1, y:y_1, x:x_1] = np.ones((inds[-1]+1-inds[0], y_1-y, x_1-x))

        # Update the local variables with remote data
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Filling cache with: Time - %s:%s, Lat - %s:%s, Lon - %s:%s" % (str(inds[0]), str(inds[-1]+1), str(y), str(y_1), str(x), str(x_1)))
        for local, remote in zip(localvars, remotevars):
            if len(shape) == 4:
                local[inds[0]:inds[-1]+1, 0:shape[1], y:y_1, x:x_1] = remote[inds[0]:inds[-1]+1,  0:shape[1], y:y_1, x:x_1]
            else:
                local[inds[0]:inds[-1]+1, y:y_1, x:x_1] = remote[inds[0]:inds[-1]+1, y:y_1, x:x_1]

    def __call__(self, active):
        c = 0

        self.dataset = CommonDataset.open(self.hydrodataset)
        self.remote = self.dataset.nc

        # Calculate the datetimes of the model timesteps like
        # the particle objects do, so we can figure out unique
        # time indices
        modelTimestep, newtimes = AsaTransport.get_time_objects_from_model_timesteps(self.times, start=self.start_time)

        timevar = self.dataset.gettimevar(self.uname)

        # Don't need to grab the last datetime, as it is not needed for forcing, only
        # for setting the time of the final particle forcing
        time_indexs = timevar.nearest_index(newtimes[0:-1], select='before')

        # Have to make sure that we get the plus 1 for the
        # linear interpolation of u,v,w,temp,salt
        self.inds = np.unique(time_indexs)
        self.inds = np.append(self.inds, self.inds.max()+1)

        # While there is at least 1 particle still running,
        # stay alive, if not break
        while self.n_run.value > 1:

            if self.caching is False:
                logger.debug("Caching is False, not doing much.  Just hanging out until all of the particles finish.")
                timer.sleep(10)
                continue

            # If particle asks for data, do the following
            if self.get_data.value is True:
                logger.debug("Particle asked for data!")

                # Wait for particles to get out
                while True:
                    self.read_lock.acquire()

                    logger.debug("Read count: %d" % self.read_count.value)
                    if self.read_count.value > 0:
                        logger.debug("Waiting for write lock on cache file (particles must stop reading)...")
                        self.read_lock.release()
                        timer.sleep(2)
                    else:
                        break

                # Get write lock on the file.  Already have read lock.
                self.write_lock.acquire()
                self.has_write_lock.value = os.getpid()

                if c == 0:
                    logger.debug("Creating cache file")
                    try:
                        # Open local cache for writing, overwrites
                        # existing file with same name
                        self.local = netCDF4.Dataset(self.cache_path, 'w')

                        indices = self.dataset.get_indices(self.uname, timeinds=[np.asarray([0])], point=self.start)
                        self.point_get.value = [self.inds[0], indices[-2], indices[-1]]

                        # Create dimensions for u and v variables
                        self.local.createDimension('time', None)
                        self.local.createDimension('level', None)
                        self.local.createDimension('x', None)
                        self.local.createDimension('y', None)

                        # Create 3d or 4d u and v variables
                        if self.remote.variables[self.uname].ndim == 4:
                            self.ndim = 4
                            dimensions = ('time', 'level', 'y', 'x')
                            coordinates = "time z lon lat"
                        elif self.remote.variables[self.uname].ndim == 3:
                            self.ndim = 3
                            dimensions = ('time', 'y', 'x')
                            coordinates = "time lon lat"
                        shape = self.remote.variables[self.uname].shape

                        # If there is no FillValue defined in the dataset, use np.nan.
                        # Sometimes it will work out correctly and other times we will
                        # have a huge cache file.
                        try:
                            fill = self.remote.variables[self.uname].missing_value
                        except Exception:
                            fill = np.nan

                        # Create domain variable that specifies
                        # where there is data geographically/by time
                        # and where there is not data,
                        #   Used for testing if particle needs to
                        #   ask cache to update
                        domain = self.local.createVariable('domain', 'i', dimensions, zlib=False, fill_value=0)
                        domain.coordinates = coordinates

                        # Create local u and v variables
                        u = self.local.createVariable('u', 'f', dimensions, zlib=False, fill_value=fill)
                        v = self.local.createVariable('v', 'f', dimensions, zlib=False, fill_value=fill)

                        v.coordinates = coordinates
                        u.coordinates = coordinates

                        localvars = [u, v, ]
                        remotevars = [self.remote.variables[self.uname], self.remote.variables[self.vname]]

                        # Create local w variable
                        if self.wname is not None:
                            w = self.local.createVariable('w', 'f', dimensions, zlib=False, fill_value=fill)
                            w.coordinates = coordinates
                            localvars.append(w)
                            remotevars.append(self.remote.variables[self.wname])

                        if self.temp_name is not None and self.salt_name is not None:
                            # Create local temp and salt vars
                            temp = self.local.createVariable('temp', 'f', dimensions, zlib=False, fill_value=fill)
                            salt = self.local.createVariable('salt', 'f', dimensions, zlib=False, fill_value=fill)
                            temp.coordinates = coordinates
                            salt.coordinates = coordinates
                            localvars.append(temp)
                            localvars.append(salt)
                            remotevars.append(self.remote.variables[self.temp_name])
                            remotevars.append(self.remote.variables[self.salt_name])

                        # Create local lat/lon coordinate variables
                        if self.remote.variables[self.xname].ndim == 2:
                            lon = self.local.createVariable('lon', 'f', ("y", "x"), zlib=False)
                            lon[:] = self.remote.variables[self.xname][:, :]
                            lat = self.local.createVariable('lat', 'f', ("y", "x"), zlib=False)
                            lat[:] = self.remote.variables[self.yname][:, :]
                        if self.remote.variables[self.xname].ndim == 1:
                            lon = self.local.createVariable('lon', 'f', ("x"), zlib=False)
                            lon[:] = self.remote.variables[self.xname][:]
                            lat = self.local.createVariable('lat', 'f', ("y"), zlib=False)
                            lat[:] = self.remote.variables[self.yname][:]

                        # Create local z variable
                        if self.zname is not None:
                            if self.remote.variables[self.zname].ndim == 4:
                                z = self.local.createVariable('z', 'f', ("time", "level", "y", "x"), zlib=False)
                                remotez = self.remote.variables[self.zname]
                                localvars.append(z)
                                remotevars.append(remotez)
                            elif self.remote.variables[self.zname].ndim == 3:
                                z = self.local.createVariable('z', 'f', ("level", "y", "x"), zlib=False)
                                z[:] = self.remote.variables[self.zname][:, :, :]
                            elif self.remote.variables[self.zname].ndim == 1:
                                z = self.local.createVariable('z', 'f', ("level",), zlib=False)
                                z[:] = self.remote.variables[self.zname][:]

                        # Create local time variable
                        time = self.local.createVariable('time', 'f8', ("time",), zlib=False)
                        if self.tname is not None:
                            time[:] = self.remote.variables[self.tname][self.inds]

                        if self.point_get.value[0]+self.time_size > np.max(self.inds):
                            current_inds = np.arange(self.point_get.value[0], np.max(self.inds)+1)
                        else:
                            current_inds = np.arange(self.point_get.value[0], self.point_get.value[0] + self.time_size)

                        # Get data from remote dataset and add
                        # to local cache.
                        # Try 20 times on the first attempt
                        current_attempt = 1
                        max_attempts = 20
                        while True:
                            try:
                                assert current_attempt <= max_attempts
                                self.get_remote_data(localvars, remotevars, current_inds, shape)
                            except AssertionError:
                                raise
                            except:
                                logger.warn("CachingDataController failed to get remote data.  Trying again in 20 seconds. %s attempts left." % str(max_attempts-current_attempt))
                                logger.exception("Data Access Error")
                                timer.sleep(20)
                                current_attempt += 1
                            else:
                                break

                        c += 1
                    except (Exception, AssertionError):
                        logger.error("CachingDataController failed to get data (first request)")
                        raise
                    finally:
                        self.local.sync()
                        self.local.close()
                        self.has_write_lock.value = -1
                        self.write_lock.release()
                        self.get_data.value = False
                        self.read_lock.release()
                        logger.debug("Done updating cache file, closing file, and releasing locks")
                else:
                    logger.debug("Updating cache file")
                    try:
                        # Open local cache dataset for appending
                        self.local = netCDF4.Dataset(self.cache_path, 'a')

                        # Create local and remote variable objects
                        # for the variables of interest
                        u = self.local.variables['u']
                        v = self.local.variables['v']
                        time = self.local.variables['time']
                        remoteu = self.remote.variables[self.uname]
                        remotev = self.remote.variables[self.vname]

                        # Create lists of variable objects for
                        # the data updater
                        localvars = [u, v, ]
                        remotevars = [remoteu, remotev, ]
                        if self.salt_name is not None and self.temp_name is not None:
                            salt = self.local.variables['salt']
                            temp = self.local.variables['temp']
                            remotesalt = self.remote.variables[self.salt_name]
                            remotetemp = self.remote.variables[self.temp_name]
                            localvars.append(salt)
                            localvars.append(temp)
                            remotevars.append(remotesalt)
                            remotevars.append(remotetemp)
                        if self.wname is not None:
                            w = self.local.variables['w']
                            remotew = self.remote.variables[self.wname]
                            localvars.append(w)
                            remotevars.append(remotew)
                        if self.zname is not None:
                            remotez = self.remote.variables[self.zname]
                            if remotez.ndim == 4:
                                z = self.local.variables['z']
                                localvars.append(z)
                                remotevars.append(remotez)
                        if self.tname is not None:
                            # remotetime = self.remote.variables[self.tname]
                            time[self.inds] = self.remote.variables[self.inds]

                        if self.point_get.value[0]+self.time_size > np.max(self.inds):
                            current_inds = np.arange(self.point_get.value[0], np.max(self.inds)+1)
                        else:
                            current_inds = np.arange(self.point_get.value[0], self.point_get.value[0] + self.time_size)

                        # Get data from remote dataset and add
                        # to local cache
                        while True:
                            try:
                                self.get_remote_data(localvars, remotevars, current_inds, shape)
                            except:
                                logger.warn("CachingDataController failed to get remote data.  Trying again in 30 seconds")
                                timer.sleep(30)
                            else:
                                break

                        c += 1
                    except Exception:
                        logger.error("CachingDataController failed to get data (not first request)")
                        raise
                    finally:
                        self.local.sync()
                        self.local.close()
                        self.has_write_lock.value = -1
                        self.write_lock.release()
                        self.get_data.value = False
                        self.read_lock.release()
                        logger.debug("Done updating cache file, closing file, and releasing locks")
            else:
                logger.debug("Particles are still running, waiting for them to request data...")
                timer.sleep(2)

        self.dataset.closenc()

        return "CachingDataController"



import math
from paegan.utils.asamath import AsaMath
from paegan.utils.asagreatcircle import AsaGreatCircle
from shapely.geometry import Point, Polygon, MultiPolygon
import random
import time
from datetime import timedelta

from paegan.logger import logger

class AsaTransport(object):

    @classmethod
    def get_time_objects_from_model_timesteps(cls, times, start):
        """
        Calculate the datetimes of the model timesteps

        times should start at 0 and be in seconds

        """
        modelTimestep = []
        newtimes = []

        for i in range(0, len(times)):
            try:
                modelTimestep.append(times[i+1] - times[i])
            except Exception:
                modelTimestep.append(times[i] - times[i-1])
            newtimes.append(start + timedelta(seconds=times[i]))
      
        return (modelTimestep, newtimes)

    @classmethod
    def fill_polygon_with_points(cls, goal=None, polygon=None):
        """
            Fill a shapely polygon with X number of points
        """
        if goal is None:
            raise ValueError("Must specify the number of points (goal) to fill the polygon with")

        if polygon is None or (not isinstance(polygon, Polygon) and not isinstance(polygon, MultiPolygon)):
            raise ValueError("Must specify a polygon to fill points with")

        minx = polygon.bounds[0] 
        maxx = polygon.bounds[2] 
        miny = polygon.bounds[1] 
        maxy = polygon.bounds[3] 

        points = []
        now = time.time()
        while len(points) < goal:
            random_x = random.uniform(minx, maxx)
            random_y = random.uniform(miny, maxy)
            p = Point(random_x, random_y)
            if p.within(polygon):
                points.append(p)

        logger.info("Filling polygon with points took %f seconds" % (time.time() - now))

        return points

    @classmethod
    def distance_from_location_using_u_v_w(cls, u=None, v=None, w=None, timestep=None, location=None):
        """
            Calculate the greate distance from a location using u, v, and w.

            u, v, and w must be in the same units as the timestep.  Stick with seconds.

        """
        # Move horizontally
        distance_horiz = 0
        azimuth = 0
        angle = 0
        depth = location.depth

        if u is not 0 and v is not 0:
            s_and_d = AsaMath.speed_direction_from_u_v(u=u,v=v) # calculates velocity in m/s from transformed u and v
            distance_horiz = s_and_d['speed'] * timestep # calculate the horizontal distance in meters using the velocity and model timestep
            angle = s_and_d['direction']
            # Great circle calculation 
            # Calculation takes in azimuth (heading from North, so convert our mathematical angle to azimuth)
            azimuth = AsaMath.math_angle_to_azimuth(angle=angle)
            
        distance_vert = 0.
        if w is not None:
            # Move vertically
            # Depth is positive up, negative down.  w wil be negative if moving down, and positive if moving up
            distance_vert = w * timestep
            depth += distance_vert # calculate the vertical distance in meters using w (m/s) and model timestep (s)
            

        if distance_horiz != 0:
            vertical_angle = math.degrees(math.atan(distance_vert / distance_horiz))
            gc_result = AsaGreatCircle.great_circle(distance=distance_horiz, azimuth=azimuth, start_point=location) 
        else:
            # Did we go up or down?
            vertical_angle = 0.
            if distance_vert < 0:
                # Down
                vertical_angle = 270.
            elif distance_vert > 0:
                # Up
                vertical_angle = 90.
            gc_result = { 'latitude': location.latitude, 'longitude': location.longitude, 'reverse_azimuth': 0 }
            
        #logger.info("Particle moving from %fm to %fm from a vertical speed of %f m/s over %s seconds" % (location.depth, depth, w, str(timestep)))            
        gc_result['azimuth'] = azimuth
        gc_result['depth'] = depth
        gc_result['distance'] = distance_horiz
        gc_result['angle'] = angle
        gc_result['vertical_distance'] = distance_vert
        gc_result['vertical_angle'] = vertical_angle
        return gc_result



try:
    from __builtin__ import unicode as str
except:
    pass

import os
import time
import pytz
import logging
import threading
import multiprocessing
import multiprocessing.pool

from shapely.geometry import Point, Polygon, MultiPolygon
from shapely.ops import cascaded_union

from paegan.transport.particles.particle import LarvaParticle
from paegan.location4d import Location4D
from paegan.transport.utils.asatransport import AsaTransport
from paegan.transport.shoreline import Shoreline
from paegan.cdm.dataset import CommonDataset
from paegan.transport.exceptions import BaseDataControllerError
from paegan.transport.forcers import BaseForcer
from paegan.logger import logger
import paegan.transport.export as ex


class Runner(object):
    """
    Shim needed for Pool/cluster runnable interface.
    """
    def __call__(self, task):
        return task(None)


class BaseModelController(object):
    """
        Controls the models
    """
    def __init__(self, **kwargs):

        """
            Mandatory named arguments:
            * geometry (Shapely Geometry Object) no default
            * depth (meters) default 0
            * start (DateTime Object) none
            * step (seconds) default 3600
            * npart (number of particles) default 1
            * nstep (number of steps) no default
            * models (list object) no default, so far there is a transport model and a behavior model
            geometry is interchangeable (if it is a point release) with:
            * latitude (DD) no default
            * longitude (DD) no default
            * depth (meters) default 0

            Non-mandatory named arguments:
            * pool (task pool) - defaults to multiprocessing.pool, inject your own for cluster ability
        """

        # Should we thread the result listener?  Defined in subclasses that need it.
        self.thread_result_listener = False

        # Shoreline
        self._use_shoreline         = kwargs.pop('use_shoreline', True)
        self.shoreline_path         = kwargs.get("shoreline_path", None)
        self.shoreline_feature      = kwargs.get("shoreline_feature", None)
        self.shoreline_index_buffer = kwargs.get("shoreline_index_buffer", 0.1)
        self.reverse_distance       = kwargs.get("reverse_distance", 100)

        # Bathy
        self._use_bathymetry = kwargs.pop('use_bathymetry', True)
        self.bathy_path      = kwargs.get("bathy_path", None)

        # SeaSurface
        self._use_seasurface = kwargs.pop('use_seasurface', True)

        self._depth          = kwargs.pop('depth', 0)
        self._npart          = kwargs.pop('npart', 1)
        self._step           = kwargs.pop('step', 3600)
        self.start           = kwargs.get('start', None)
        if self.start is None:
            raise TypeError("must provide a start time to run the model")

        # Always convert to UTC
        if self.start.tzinfo is None:
            self.start = self.start.replace(tzinfo=pytz.utc)
        self.start = self.start.astimezone(pytz.utc)

        self._models = kwargs.pop('models', None)
        self._dirty  = True

        self.particles              = []
        self.time_method            = kwargs.get('time_method', 'interp').lower()
        try:
            assert "interp" == self.time_method or "nearest" == self.time_method
        except:
            raise TypeError("Not a recognized 'time_method' parameter.  Only 'nearest' or 'interp' are allowed.")

        # The model timesteps in datetime objects
        self.datetimes = []

        # Interchangeables
        if "geometry" in kwargs:
            self.geometry = kwargs.pop('geometry')
            if not isinstance(self.geometry, Point) and not isinstance(self.geometry, Polygon) and not isinstance(self.geometry, MultiPolygon):
                raise TypeError("The geometry attribute must be a shapely Point or Polygon")
        elif "latitude" and "longitude" in kwargs:
            self.geometry = Point(kwargs.pop('longitude'), kwargs.pop('latitude'))
        else:
            raise TypeError("must provide a shapely geometry object (point or polygon) or a latitude and a longitude")

        # Errors
        try:
            self._nstep = kwargs.pop('nstep')
        except Exception:
            logger.exception("Must provide the number of timesteps to the ModelController")
            raise

        self.pool = kwargs.get('pool', None)

    def set_geometry(self, geo):
        # If polygon is passed in, we need to trim it by the coastline
        # so we don't start particles on land
        if isinstance(geo, Polygon) and self._use_shoreline:
            c = geo.centroid
            b = geo.bounds
            spatialbuffer = max(b[2] - b[0], b[3] - b[1])
            shore_geoms = Shoreline(path=self.shoreline_path, feature_name=self.shoreline_feature, point=c, spatialbuffer=spatialbuffer).geoms
            if len(shore_geoms) > 0:
                all_shore = cascaded_union(shore_geoms)
                geo = geo.difference(all_shore)

        self._geometry = geo

    def get_geometry(self):
        return self._geometry
    geometry = property(get_geometry, set_geometry)

    def get_reference_location(self):
        pt = self.geometry.centroid
        return Location4D(latitude=pt.y, longitude=pt.x, depth=self._depth, time=self.start)
    reference_location = property(get_reference_location, None)

    def set_start(self, sta):
        self._start = sta

    def get_start(self):
        return self._start
    start = property(get_start, set_start)

    def set_particles(self, parts):
        self._particles = parts

    def get_particles(self):
        return self._particles
    particles = property(get_particles, set_particles)

    def __str__(self):
        return "*** BaseModelController ***"

    def get_common_variables_from_dataset(self, dataset):

        def getname(names):
            for n in names:
                nm = dataset.get_varname_from_stdname(n)
                if len(nm) > 0:
                    return nm[0]
                else:
                    continue
            return None

        uname = getname(['eastward_sea_water_velocity', 'eastward_current'])
        vname = getname(['northward_sea_water_velocity', 'northward_current'])
        wname = getname('upward_sea_water_velocity')
        temp_name = getname('sea_water_temperature')
        salt_name = getname('sea_water_salinity')

        coords = dataset.get_coord_names(uname)
        xname = coords['xname']
        yname = coords['yname']
        zname = coords['zname']
        tname = coords['tname']
        tname = None  # temporary

        return {
            "u"     :   uname,
            "v"     :   vname,
            "w"     :   wname,
            "temp"  :   temp_name,
            "salt"  :   salt_name,
            "x"     :   xname,
            "y"     :   yname,
            "z"     :   zname,
            "time"  :   tname
        }

    def total_particle_count(self):
        return len(self.particles)

    def total_task_count(self):
        return self.total_particle_count()

    def listen_for_results(self, output_h5_file, total_particles):
        logger.info("Waiting for %i particle results" % total_particles)

        particles = []
        retrieved = 0
        timeout = 200

        while retrieved < total_particles:
            try:
                # self.result is an iterator that can timeout on next()
                particle = self.result.next(timeout)
                retrieved += 1
                particles.append(particle)
            except StopIteration:
                assert retrieved >= total_particles
                break
            except:
                logger.exception("Particle has FAILED!!")
                continue

            # We multiply by 90 here to save 10% for the exporting
            logger.progress((round((float(retrieved) / total_particles) * 90., 1), "%s Particle(s) complete" % retrieved))

        logger.info(particles)
        results = ex.ResultsPyTable(output_h5_file)
        for p in particles:
            for x in range(len(p.locations)):
                results.write(p.timestep_index_dump(x))
        results.compute()
        results.close()

        return

    def cleanup(self):

        # Remove timevar
        del self.timevar

    def start_tasks(self, **kwargs):

        if self.pool is None:
            self.pool = multiprocessing.Pool()

        try:
            logger.info('Adding %i particles as tasks' % self.total_particle_count())
            tasks = []

            for part in self.particles:
                forcer = BaseForcer(self.hydrodataset,
                                    particle=part,
                                    common_variables=self.common_variables,
                                    timevar=self.timevar,
                                    times=self.times,
                                    start_time=self.start,
                                    models=self._models,
                                    release_location_centroid=self.reference_location.point,
                                    usebathy=self._use_bathymetry,
                                    useshore=self._use_shoreline,
                                    usesurface=self._use_seasurface,
                                    reverse_distance=self.reverse_distance,
                                    bathy_path=self.bathy_path,
                                    shoreline_path=self.shoreline_path,
                                    shoreline_feature=self.shoreline_feature,
                                    time_method=self.time_method,
                                    shoreline_index_buffer=self.shoreline_index_buffer)
                tasks.append(forcer)

            logger.progress((5, 'Running model'))
            return self.pool.imap(Runner(), tasks)

        except Exception:
            logger.exception("Something didn't start correctly!")
            raise

    def setup_run(self, hydrodataset, **kwargs):

        self.hydrodataset = hydrodataset

        logger.setLevel(logging.PROGRESS)

        # Relax.
        time.sleep(0.5)

        # Add ModelController description to logfile
        logger.info(str(self))

        # Add the model descriptions to logfile
        for m in self._models:
            logger.info(str(m))

        # Calculate the model timesteps
        # We need times = len(self._nstep) + 1 since data is stored one timestep
        # after a particle is forced with the final timestep's data.
        self.times = list(range(0, (self._step*self._nstep)+1, self._step))
        # Calculate a datetime object for each model timestep
        # This method is duplicated in CachingDataController and CachingForcer
        # using the 'times' variables above.  Will be useful in those other
        # locations for particles released at different times
        # i.e. released over a few days
        self.modelTimestep, self.datetimes = AsaTransport.get_time_objects_from_model_timesteps(self.times, start=self.start)

        logger.progress((1, "Setting up particle start locations"))
        point_locations = []
        if isinstance(self.geometry, Point):
            point_locations = [self.reference_location] * self._npart
        elif isinstance(self.geometry, Polygon) or isinstance(self.geometry, MultiPolygon):
            point_locations = [Location4D(latitude=loc.y, longitude=loc.x, depth=self._depth, time=self.start) for loc in AsaTransport.fill_polygon_with_points(goal=self._npart, polygon=self.geometry)]

        # Initialize the particles
        logger.progress((2, "Initializing particles"))
        for x in range(0, self._npart):
            p = LarvaParticle(id=x)
            p.location = point_locations[x]
            # We don't need to fill the location gaps here for environment variables
            # because the first data collected actually relates to this original
            # position.
            # We do need to fill in fields such as settled, halted, etc.
            p.fill_status_gap()
            # Set the inital note
            p.note = p.outputstring()
            p.notes.append(p.note)
            self.particles.append(p)

        logger.progress((3, "Initializing and caching hydro model's grid %s" % self.hydrodataset))
        try:
            ds = CommonDataset.open(self.hydrodataset)
            # Query the dataset for common variable names
            # and the time variable.
            logger.debug("Retrieving variable information from dataset")
            self.common_variables = self.get_common_variables_from_dataset(ds)
        except Exception:
            logger.exception("Failed to access dataset %s" % self.hydrodataset)
            raise BaseDataControllerError("Inaccessible Dataset: %s" % self.hydrodataset)

        self.timevar = None
        try:
            assert self.common_variables.get("u") in ds._current_variables
            assert self.common_variables.get("v") in ds._current_variables
            assert self.common_variables.get("x") in ds._current_variables
            assert self.common_variables.get("y") in ds._current_variables

            self.timevar = ds.gettimevar(self.common_variables.get("u"))
            model_start = self.timevar.get_dates()[0]
            model_end = self.timevar.get_dates()[-1]
        except AssertionError:
            logger.exception("Could not locate variables needed to run model: %s" % str(self.common_variables))
            raise BaseDataControllerError("A required data variable was not found in %s" % self.hydrodataset)
        finally:
            ds.closenc()

        try:
            assert self.start > model_start
            assert self.start < model_end
        except AssertionError:
            raise BaseDataControllerError("Start time for model (%s) is not available in source dataset (%s/%s)" % (self.datetimes[0], model_start, model_end))

        try:
            assert self.datetimes[-1] > model_start
            assert self.datetimes[-1] < model_end
        except AssertionError:
            raise BaseDataControllerError("End time for model (%s) is not available in source dataset (%s/%s)" % (self.datetimes[-1], model_start, model_end))

    def run(self, **kwargs):

        logger.progress((4, "Starting tasks"))
        self.result = self.start_tasks(**kwargs)
        if self.result is None:
            raise BaseDataControllerError("Not all tasks started! Exiting.")

        # Store results in hdf5 file for processing later
        output_h5_file = None
        if kwargs.get('output_path') is not None:
            output_h5_file = os.path.join(kwargs.get('output_path'), 'results.h5')

        if self.thread_result_listener is True:
            rl = threading.Thread(name="ResultListener", target=self.listen_for_results, args=(output_h5_file, self.total_particle_count()))
            rl.daemon = True
            rl.start()
            rl.join()  # This blocks until the tasks are all done.
        else:
            self.listen_for_results(output_h5_file, self.total_particle_count())    # This blocks until the tasks are all done.

        logger.info('Tasks are all finished... Cleaning up!!')
        self.cleanup()

        # If output_formats and path specified,
        # output particle run data to disk when completed
        if "output_formats" in kwargs:

            logger.progress((96, "Exporting results"))

            # Make sure output_path is also included
            if kwargs.get("output_path", None) is not None:
                formats = kwargs.get("output_formats")
                output_path = kwargs.get("output_path")
                if isinstance(formats, list):
                    for fmt in formats:
                        logger.info("Exporting to: %s" % fmt)
                        try:
                            # Calls the export function
                            fmt.export(output_path, output_h5_file)
                        except:
                            logger.exception("Failed to export to: %s" % fmt)
                else:
                    logger.warn('The output_formats parameter should be a list, not saving any output!')
            else:
                logger.warn('No output path defined, not saving any output!')
        else:
            logger.warn('No output_formats parameter was defined, not saving any output!')

        logger.progress((97, "Model Run Complete"))

        return



try:
    from __builtin__ import unicode as str
except ImportError:
    pass

import calendar
from shapely.geometry import LineString


def or_none(func):
    try:
        return func()
    except BaseException:
        return None


def index_or_none(lst, index):
    try:
        return lst[index]
    except IndexError:
        return None


class Particle(object):
    """
        A particle
    """
    def __init__(self, **kwargs):
        self._id = kwargs.get("id", -1)
        self._locations = []

        self._age = 0. # Age in days
        self._ages = [0.] # Ages in days

        self._u = None
        self._us = []

        self._v = None
        self._vs = []

        self._w = None
        self._ws = []

        self._halted = False
        self._halts = []

        self._note = ""
        self._notes = []

    def get_id(self):
        return self._id
    uid = property(get_id)

    def set_location(self, location):
        self._locations.append(location)
    def get_location(self):
        return self._locations[-1]
    location = property(get_location, set_location)
    def get_locations(self):
        return self._locations
    locations = property(get_locations, None)

    def save(self):
        self.halts.append(self.halted)
        self.u_vectors.append(self.u_vector)
        self.v_vectors.append(self.v_vector)
        self.w_vectors.append(self.w_vector)
        self.notes.append(self.note)
        self.ages.append(self.get_age(units='days'))

        self.u_vector = None
        self.v_vector = None
        self.w_vector = None
        self.note = ""

    def fill_status_gap(self, value=None):

        def fillvalue(v):
            if str(value).lower() == u"last":
                return v
            else:
                return value

        if len(self.locations) > len(self.halts):
            self.halts.append(self.halted)

        if len(self.locations) > len(self.ages):
            self.ages.append(self.get_age(units='days'))

    def fill_environment_gap(self, value=None):

        def fillvalue(v):
            if str(value).lower() == u"last":
                return v
            else:
                return value

        if len(self.locations) > len(self.u_vectors):
            self.u_vectors.append(fillvalue(self.last_u()))

        if len(self.locations) > len(self.v_vectors):
            self.v_vectors.append(fillvalue(self.last_v()))

        if len(self.locations) > len(self.w_vectors):
            self.w_vectors.append(fillvalue(self.last_w()))

    def fill_gap(self, value=None):
        self.fill_environment_gap(value=value)
        self.fill_status_gap(value=value)

    def set_u_vector(self, u):
        self._u = u
    def get_u_vector(self):
        return self._u
    u_vector = property(get_u_vector, set_u_vector)
    def get_u_vectors(self):
        return self._us
    u_vectors = property(get_u_vectors, None)
    def last_u(self):
        try:
            return float(self.u_vectors[-1])
        except (IndexError, TypeError):
            return None

    def set_v_vector(self, v):
        self._v = v
    def get_v_vector(self):
        return self._v
    v_vector = property(get_v_vector, set_v_vector)
    def get_v_vectors(self):
        return self._vs
    v_vectors = property(get_v_vectors, None)
    def last_v(self):
        try:
            return float(self.v_vectors[-1])
        except (IndexError, TypeError):
            return None

    def set_w_vector(self, w):
        self._w = w
    def get_w_vector(self):
        return self._w
    w_vector = property(get_w_vector, set_w_vector)
    def get_w_vectors(self):
        return self._ws
    w_vectors = property(get_w_vectors, None)
    def last_w(self):
        try:
            return float(self.w_vectors[-1])
        except (IndexError, TypeError):
            return None

    def get_ages(self):
        return self._ages
    ages = property(get_ages, None)

    def set_note(self, note):
        self._note = note
    def get_note(self):
        return self._note
    note = property(get_note, set_note)
    def get_notes(self):
        return self._notes
    notes = property(get_notes, None)
    def last_note(self):
        try:
            return self.notes[-1]
        except IndexError:
            return ""

    def proceed(self):
        """
        Unhalt the movement of this particle.
        """
        self._halted = False
    def halt(self):
        """
        Halt the movement of this particle.
        """
        self._halted = True
    def get_halted(self):
        return self._halted
    halted = property(get_halted, None)
    def get_halts(self):
        return self._halts
    halts = property(get_halts, None)

    def set_active(self, active):
        self._active = active
    def get_active(self):
        return self._active
    active = property(get_active, set_active)

    def get_last_movement(self):
        return LineString(list(self.locations[-2].point.coords) + list(self.locations[-1].point.coords))

    def get_age(self, **kwargs):
        """
        Returns the particlees age (how long it has been forced) in a variety of units.
        Rounded to 8 decimal places.

        Parameters:
            units (optional) = 'days' (default), 'hours', 'minutes', or 'seconds'
        """
        try:
            units = kwargs.get('units', None)
            if units is None:
                return self._age
            units = units.lower()
            if units == "days":
                z = self._age
            elif units == "hours":
                z = self._age * 24.
            elif units == "minutes":
                z = self._age * 24. * 60.
            elif units == "seconds":
                z = self._age * 24. * 60. * 60.
            else:
                raise
            return round(z,8)
        except Exception:
            raise KeyError("Could not return age of particle")

    def age(self, **kwargs):
        """
        Age this particle.

        parameters (optional, only one allowed):
            days (default)
            hours
            minutes
            seconds
        """
        if kwargs.get('days', None) is not None:
            self._age += kwargs.get('days')
        elif kwargs.get('hours', None) is not None:
            self._age += kwargs.get('hours') / 24.
        elif kwargs.get('minutes', None) is not None:
            self._age += kwargs.get('minutes') / 24. / 60.
        elif kwargs.get('seconds', None) is not None:
            self._age += kwargs.get('seconds') / 24. / 60. / 60.
        else:
            raise KeyError("Could not age particle, please specify 'days', 'hours', 'minutes', or 'seconds' parameter")
        return

    def linestring(self):
        return LineString([list(x.point.coords)[0] for x in self.locations])

    def normalized_locations(self, model_timesteps):
        inds = self.normalized_indexes(model_timesteps)
        return [loc for i,loc in enumerate(self.locations) if i in inds]

    def normalized_indexes(self, model_timesteps):
        """
        This function will normalize the particles locations
        to the timestep of the model that was run.  This is used
        in output, as we should only be outputting the model timestep
        that was chosen to be run.

        In most cases, the length of the model_timesteps and the
        particle's locations will be the same (unless it hits shore).

        If they are not the same length pull out of locations the timesteps
        that are closest to the model_timesteps
        """

        # Clean up locations
        # If duplicate time instances, remove the lower index
        clean_locs = []
        for i,loc in enumerate(self.locations):
            try:
                if loc.time == self.locations[i+1].time:
                    continue
                else:
                    clean_locs.append(loc)
            except Exception:
                clean_locs.append(loc)

        if len(clean_locs) == len(model_timesteps):
           return [ind for ind,loc in enumerate(self.locations) if loc in clean_locs]
        elif len(model_timesteps) < len(clean_locs):
            # We have at least one internal timestep for this particle
            # Pull out the matching location indexes
            indexes = [ind for ind,loc in enumerate(self.locations) if loc in clean_locs]
            if len(model_timesteps) == len(indexes):
                return indexes
            raise ValueError("Can't normalize")
        elif len(model_timesteps) > len(clean_locs):
            # The particle stopped before forcing for all of the model timesteps
            raise ValueError("Particle has less locations than model timesteps")

    def timestep_dump(self):
        return { "particle"  : self.uid,
                 "u_vector"  : self.last_u(),
                 "v_vector"  : self.last_v(),
                 "w_vector"  : self.last_w(),
                 "time"      : calendar.timegm(self.location.time.utctimetuple()),
                 "latitude"  : self.location.latitude,
                 "longitude" : self.location.longitude,
                 "depth"     : self.location.depth,
                 "halted"    : self.halted,
                 "age"       : self.get_age(units='days') }

    def timestep_index_dump(self, index):

        loc = index_or_none(self.locations, index)

        return { "particle"  : self.uid,
                 "u_vector"  : index_or_none(self.u_vectors, index),
                 "v_vector"  : index_or_none(self.v_vectors, index),
                 "w_vector"  : index_or_none(self.w_vectors, index),
                 "time"      : calendar.timegm(loc.time.utctimetuple()) if loc else None,
                 "latitude"  : loc.latitude if loc else None,
                 "longitude" : loc.longitude if loc else None,
                 "depth"     : loc.depth if loc else None,
                 "halted"    : index_or_none(self.halts, index),
                 "age"       : index_or_none(self.ages, index)}


class LarvaParticle(Particle):
    """
        A particle for larvamap, keeps track of information
        for the behaviors component
    """
    def __init__(self, **kwargs):
        super(LarvaParticle,self).__init__(**kwargs)
        self.lifestage_progress = 0.
        self.lifestage_progresses = [0.]

        self._temp = None
        self._temps = []

        self._salt = None
        self._salts = []

        self._settled = False
        self._settles = []

        self._dead = False
        self._deads = []

    # Temp
    def set_temp(self, temp):
        self._temp = temp
    def get_temp(self):
        return self._temp
    temp = property(get_temp, set_temp)
    def get_temps(self):
        return self._temps
    temps = property(get_temps, None)
    def last_temp(self):
        try:
            return float(self.temps[-1])
        except (IndexError, TypeError):
            return None

    # Salt
    def set_salt(self, salt):
        self._salt = salt
    def get_salt(self):
        return self._salt
    salt = property(get_salt, set_salt)
    def get_salts(self):
        return self._salts
    salts = property(get_salts, None)
    def last_salt(self):
        try:
            return float(self.salts[-1])
        except (IndexError, TypeError):
            return None

    def get_lifestage_index(self):
        return int(self.lifestage_progress)
    lifestage_index = property(get_lifestage_index, None)

    def settle(self):
        self._settled = True
        self.halt()
    def unsettle(self):
        self._settled = False
        self.proceed()
    def get_settled(self):
        return self._settled
    settled = property(get_settled, None)
    def get_settles(self):
        return self._settles
    settles = property(get_settles, None)

    def die(self):
        self._dead = True
        self._deads
    def undie(self):
        self._dead = False
    def get_dead(self):
        return self._dead
    dead = property(get_dead, None)
    def get_deads(self):
        return self._deads
    deads = property(get_deads, None)

    def save(self):
        super(LarvaParticle,self).save()
        self.deads.append(self.dead)
        self.settles.append(self.settled)
        self.temps.append(self.temp)
        self.salts.append(self.salt)

        self.ages.append(self.get_age(units='days'))

        self.temp = None
        self.salt = None

    def fill_status_gap(self, value=None):
        super(LarvaParticle,self).fill_status_gap(value=value)

        def fillvalue(v):
            if str(value).lower() == u"last":
                return v
            else:
                return value

        if len(self.locations) > len(self.deads):
            self.deads.append(self.dead)

        if len(self.locations) > len(self.settles):
            self.settles.append(self.settled)

    def fill_environment_gap(self, value=None):
        super(LarvaParticle,self).fill_environment_gap(value=value)

        def fillvalue(v):
            if str(value).lower() == u"last":
                return v
            else:
                return value

        if len(self.locations) > len(self.salts):
            self.salts.append(fillvalue(self.last_salt()))

        if len(self.locations) > len(self.temps):
            self.temps.append(fillvalue(self.last_temp()))

    def fill_gap(self, value=None):
        super(LarvaParticle,self).fill_gap(value=value)

    def grow(self, amount):
        """
        Grow a particle by a percentage value (0 < x < 1)

        When a particle grows past 1, its current lifestage is
        complete and it moves onto the next.

        The calculation to get the current lifestage index is in get_lifestage_index()
        """
        self.lifestage_progress += amount
        self.lifestage_progresses.append(self.lifestage_progress)

    def logstring(self):
        return "Particle %d (Age: %.3f days, Lifestage %d: %.3f%%, Status: %s)" % (self.uid, self.get_age(units='days'), self.lifestage_index, (self.lifestage_progress % 1) * 100., self.status())

    def outputstring(self):
        """ For shapefiles, max 254 characters """
        return "Age: %.3f days\nLifestage %d: %.3f%%\n" % (self.get_age(units='days'), self.lifestage_index, (self.lifestage_progress % 1) * 100.)

    def status(self):
        return "settled - %s / dead - %s / halted - %s" % (self.settled, self.dead, self.halted)

    def timestep_dump(self):
        master = { "temperature" : self.last_temp(),
                   "salinity"    : self.last_salt(),
                   "dead"        : self.dead,
                   "settled"     : self.settled,
                   "lifestage"   : self.lifestage_index,
                   "progress"    : (self.lifestage_progress % 1) * 100 }
        master.update(super(LarvaParticle, self).timestep_dump())
        return master

    def timestep_index_dump(self, index):

        try:
            ls = int(self.lifestage_progresses[index])
            lp = (self.lifestage_progresses[index] % 1) * 100
        except IndexError:
            ls = None
            lp = None

        master = { "temperature" : index_or_none(self.temps, index),
                   "salinity"    : index_or_none(self.salts, index),
                   "dead"        : index_or_none(self.deads, index),
                   "settled"     : index_or_none(self.settles, index),
                   "lifestage"   : ls,
                   "progress"    : lp }
        master.update(super(LarvaParticle, self).timestep_index_dump(index))
        return master


class ChemistryParticle(Particle):
    """
        A chemical particle for time and chemistry dependent
        tracers
    """
    def __init__(self):
        pass


class OilParticle(Particle):
    """
        An oil particle for special oil physics and chemistry
    """
    def __init__(self):
        pass



import json
from paegan.transport.models.base_model import BaseModel
from paegan.transport.models.behaviors.lifestage import LifeStage, DeadLifeStage

class LarvaBehavior(BaseModel):

    def __init__(self, **kwargs):

        if 'json' in kwargs or 'data' in kwargs:
            data = {}
            try:
                data = json.loads(kwargs['json'])
            except Exception:
                try:
                    data = kwargs.get('data')
                except Exception:
                    pass

        self.lifestages = []
        if data.get('lifestages', None) is not None:
            self.lifestages = [LifeStage(data=ls) for ls in data.get('lifestages')]

        # What do to if the model continues but all LifeStages have passed??
        # Be dead.
        self.lifestages.append(DeadLifeStage())

    def move(self, particle, u, v, w, modelTimestep, **kwargs):

        try:
            lifestage = self.lifestages[particle.lifestage_index]
        except IndexError:
            # The particle should never progress outside of the available lifestages because a Dead
            # particle should not progress and the last lifestage is the DeadLifeStage.
            raise

        return lifestage.move(particle, u, v, w, modelTimestep, **kwargs)

    def __str__(self):
        return """
            ***** Behavior *****
              lifestages: %d
            """ % (len(self.lifestages))



try:
    from __builtin__ import unicode as str
except ImportError:
    pass

import json
from paegan.transport.models.base_model import BaseModel

class Taxis(BaseModel):

    def __init__(self, **kwargs):

        if 'json' in kwargs or 'data' in kwargs:
            data = {}
            try:
                data = json.loads(kwargs['json'])
            except Exception:
                try:
                    data = kwargs.get('data')
                except Exception:
                    pass

            # Convert from positive down to negative down
            self.min_value = data.get('min') * -1.
            self.max_value = data.get('max') * -1.
            self.gradient = data.get('gradient',None)
            self.variable = data.get('variable',None)
            self.units = data.get('units',None)

    def get_min_value(self):
        return self._min_value
    def set_min_value(self, min_value):
        self._min_value = float(min_value)
    min_value = property(get_min_value, set_min_value)

    def get_max_value(self):
        return self._max_value
    def set_max_value(self, max_value):
        self._max_value = float(max_value)
    max_value = property(get_max_value, set_max_value)

    def get_gradient(self):
        return self._gradient
    def set_gradient(self, gradient):
        self._gradient = float(gradient)
    gradient = property(get_gradient, set_gradient)

    def get_variable(self):
        return self._variable
    def set_variable(self, variable):
        self._variable = variable
    variable = property(get_variable, set_variable)

    def get_units(self):
        return self._units
    def set_units(self, units):
        self._units = str(units)
    units = property(get_units, set_units)

    def __str__(self):
        return \
        """
          Min: %d
          Max: %d
          Gradient: %d
          Variables: %s
          Units: %s
        """ % (self.min_value, self.max_value, self.gradient, self.variable, self.units)

    def move(self, particle, u, v, w, modelTimestep, **kwargs):

        # If the particle is settled, don't move it anywhere
        if particle.settled:
            return { 'u': 0, 'v': 0, 'w': 0 }

        # If the particle is halted (but not settled), don't move it anywhere
        if particle.halted:
            return { 'u': 0, 'v': 0, 'w': 0 }

        u = 0
        v = 0
        w = 0
        return { 'u': u, 'v': v, 'w': w }



class BaseModel(object):

    def move(self, **kwargs):
        raise NotImplementedError("Must define a 'move' method on the model being called")

    @property
    def name(self):
        if hasattr(self, '_name') and self._name:
            return self._name

        return self.__class__.__name__

    @name.setter
    def name(self, value):
        self._name = value




from paegan.transport.controllers.base import BaseModelController, Runner
from paegan.transport.controllers.cached import CachingModelController
from paegan.transport.controllers.ipython import IPythonClusterModelController
from paegan.transport.controllers.distributed import DistributedModelController



import os
import math
import numpy as np
import requests
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse
from xml.etree import ElementTree as ET

import fiona

from shapely import wkt
from shapely.geometry import asShape, box
from shapely.geometry import LineString
from shapely.geometry import Point, Polygon
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from paegan.utils.asagreatcircle import AsaGreatCircle
from paegan.utils.asamath import AsaMath
from paegan.utils.asarandom import AsaRandom
from paegan.location4d import Location4D

from paegan.logger import logger


class Shoreline(object):
    """
    Base class for Shoreline.

    Uses a factory pattern:
        pass path as kwarg to have this method analyze and create appropriate
            Shoreline derived class
        pass wfs_server/feature_name as keyword args and get a ShorelineWFS back
        pass file (or nothing!) as a keyword arg and get a ShorelineFile back
    """
    def __new__(cls, **kwargs):
        if 'path' in kwargs and kwargs['path'] is not None:
            parsed = urlparse(kwargs['path'])
            if parsed.scheme.startswith('http') and parsed.netloc:
                return object.__new__(ShorelineWFS)

        if 'wfs_server' in kwargs:
            return object.__new__(ShorelineWFS)

        return object.__new__(ShorelineFile)

    def __init__(self, **kwargs):
        self._spatialbuffer = kwargs.pop("spatialbuffer", 1)
        self._type          = kwargs.pop("type", "reverse")
        point               = kwargs.pop("point", None)

        self.index(point=point)

    def close(self):
        pass

    @property
    def geoms(self):
        return self._geoms

    @property
    def linestring(self):
        points = []
        for poly in self._geoms:
            plines = list(poly.exterior.coords)
            for i in range(0, len(plines)-1):
                points.append(Point(plines[i], plines[i+1]))

            points.append(Point(np.nan, np.nan))  # blank point needed to remove crossing of lines
        return LineString([list(x.coords)[0] for x in points])

    def get_capabilities(self):
        """
        Gets capabilities.

        Queries WFS server or file for its capabilities (or simulated capabilities).
        """
        return None

    def get_feature_type_info(self):
        """
        Gets FeatureType as a python dict.

        Transforms feature_name info into python dict.
        """
        return None

    def get_geoms_for_bounds(self, bounds):
        """
        Helper method to get geometries withiin a certain bounds.
        """
        pass

    def index(self, point=None, spatialbuffer=None):
        """
            This queries the shapefile around a buffer of a point
            The results of this spatial query are used for shoreline detection.

            Using the entire shapefile without the spatial query takes over
            30 times the time with world land polygons.

        """
        pass

    def intersect(self, **kwargs):
        """
            Intersect a Line or Point Collection and the Shoreline

            Returns the point of intersection along the coastline
            Should also return a linestring buffer around the interseciton point
            so we can calculate the direction to bounce a particle.
        """
        ls = None

        if "linestring" in kwargs:
            ls = kwargs.pop('linestring')
            spoint = Point(ls.coords[0])
            epoint = Point(ls.coords[-1])
        elif "start_point" and "end_point" in kwargs:
            spoint = kwargs.get('start_point')
            epoint = kwargs.get('end_point')
            ls = LineString(list(spoint.coords) + list(epoint.coords))
        elif "single_point" in kwargs:
            spoint = kwargs.get('single_point')
            epoint = None
            ls = LineString(list(spoint.coords) + list(spoint.coords))
        else:
            raise TypeError( "must provide a LineString geometry object, (2) Point geometry objects, or (1) Point geometry object" )

        inter = False

        # If the current point lies outside of our current shapefile index,
        # re-query the shapefile in a buffer around this point
        #logger.info("Spatial query object: %s" % unicode(self._spatial_query_object))

        if self._spatial_query_object is None:
            # Index if we have not
            self.index(point=spoint)
        elif self._spatial_query_object.contains(ls):
            # If we contain the entire linestring, we have nothing more to index
            pass
        elif epoint and not self._spatial_query_object.contains(epoint):
            # If we don't contain the end point, index on that
            self.index(point=epoint)
        elif not self._spatial_query_object.contains(spoint):
            # Index on the start point
            self.index(point=spoint)

        for element in self._geoms:

            # Test if starting on land
            if element.contains(spoint):
                if epoint is None:
                    # If we only passed in one point, return the intersection is true.
                    return {'point': spoint, 'feature': None}
                else:
                    # If we are testing a linestring, raise an exception that we started on land.
                    raise Exception('Intersection starting point on land: %s %s %s' % (spoint.envelope, epoint.envelope, element.envelope))
            else:
                # If we are just checking a single point, continue looping.
                if epoint is None:
                    continue

            inter = ls.intersection(element)
            if not inter.is_empty:
                # Return the first point in the linestring, and the linestring that it hit
                if isinstance(inter, MultiLineString):
                    inter = inter.geoms[0]

                inter = Point(inter.coords[0])
                smaller_int = inter.buffer(self._spatialbuffer)
                shorelines = element.exterior.intersection(smaller_int)
                if isinstance(shorelines, LineString):
                    shorelines = [shorelines]
                else:
                    shorelines = list(shorelines)

                for shore_segment in shorelines:
                    # Once we find the linestring in the Polygon that was
                    # intersected, break out and return
                    if ls.touches(shore_segment):
                        break

                return {'point': Point(inter.x, inter.y, 0), 'feature': shore_segment or None}

        return None

    def react(self, **kwargs):
        """
            Bounce off of a shoreline
            feature = Linestring of two points, being the line segment the particle hit.
            angle = decimal degrees from 0 (x-axis), couter-clockwise (math style)
        """
        if self._type == "bounce":
            logger.warn("This shoreline type is NOT SUPPORTED and is broken")
            return self.__bounce(**kwargs)
        elif self._type == "reverse":
            return self.__reverse(**kwargs)
        else:
            logger.warn("Not reacting to shoreline (sticky with inifinite concentration)")
            return kwargs.get('hit_point')

    def __bounce(self, **kwargs):
        """
            Bounce off of the shoreline.

            NOTE: This does not work, but left here for future implementation

            feature = Linestring of two points, being the line segment the particle hit.
            angle = decimal degrees from 0 (x-axis), couter-clockwise (math style)
        """
        start_point = kwargs.pop('start_point')
        hit_point = kwargs.pop('hit_point')
        end_point = kwargs.pop('end_point')
        feature = kwargs.pop('feature')
        distance = kwargs.pop('distance')
        angle = kwargs.pop('angle')

        # Figure out the angle of the shoreline here (beta)
        points_in_shore = [Point(x) for x in list(feature.coords)]
        points_in_shore = sorted(points_in_shore, key=lambda x: x.x)

        # The point on the left (least longitude is always the first Point)
        first_shore = points_in_shore[0]
        last_shore = points_in_shore[-1]

        shoreline_x = abs(abs(first_shore.x) - abs(last_shore.x))
        shoreline_y = abs(abs(first_shore.y) - abs(last_shore.y))
        beta = math.degrees(math.atan(shoreline_x / shoreline_y))

        theta = 90 - angle - beta
        bounce_azimuth = AsaMath.math_angle_to_azimuth(angle=2 * theta + angle)

        #print "Beta:           " + str(beta)
        #print "Incoming Angle: " + str(angle)
        #print "ShorelineAngle: " + str(theta + angle)
        #print "Bounce Azimuth: " + str(bounce_azimuth)
        #print "Bounce Angle:   " + str(AsaMath.azimuth_to_math_angle(azimuth=bounce_azimuth))

        after_distance = distance - AsaGreatCircle.great_distance(start_point=start_point, end_point=hit_point)['distance']

        new_point = AsaGreatCircle.great_circle(distance=after_distance, azimuth=bounce_azimuth, start_point=hit_point)
        return Location4D(latitude=new_point['latitude'], longitude=new_point['longitude'], depth=start_point.depth)

    def __reverse(self, **kwargs):
        """
            Reverse particle just off of the shore in the direction that it came in.
            Adds a slight random factor to the distance and angle it is reversed in.
        """

        #st = time.clock()

        start_point = kwargs.pop('start_point')
        hit_point = kwargs.pop('hit_point')
        reverse_azimuth = kwargs.pop('reverse_azimuth')
        reverse_distance = kwargs.get('reverse_distance', None)
        if reverse_distance is None:
            reverse_distance = 100

        # Randomize the reverse angle slightly (+/- 5 degrees)
        random_azimuth = reverse_azimuth + AsaRandom.random() * 5

        count = 0
        nudge_distance = 0.01
        nudge_point = AsaGreatCircle.great_circle(distance=nudge_distance, azimuth=reverse_azimuth, start_point=hit_point)
        nudge_loc = Location4D(latitude=nudge_point['latitude'], longitude=nudge_point['longitude'], depth=start_point.depth)

        # Find point just offshore to do testing with.  Try 15 times (~350m).  This makes sure the start_point is in the water
        # for the next call to intersect (next while loop).
        while self.intersect(single_point=nudge_loc.point) and count < 16:
            nudge_distance *= 2
            nudge_point = AsaGreatCircle.great_circle(distance=nudge_distance, azimuth=reverse_azimuth, start_point=hit_point)
            nudge_loc = Location4D(latitude=nudge_point['latitude'], longitude=nudge_point['longitude'], depth=start_point.depth)
            count += 1

        # We tried 16 times and couldn't find a point.  This should totally never happen.
        if count == 16:
            logger.warn("LOOK: Could not find location in water to do shoreline calculation with.  Assuming particle did not move from original location")
            return start_point

        # Keep trying to throw particle back, halfing the distance each time until it is in water.
        # Only half it 6 times before giving up and returning the point which the particle came from.
        count = 0
        # Distance amount to half each iteration
        changing_distance = reverse_distance
        new_point = AsaGreatCircle.great_circle(distance=reverse_distance, azimuth=random_azimuth, start_point=hit_point)
        new_loc = Location4D(latitude=new_point['latitude'], longitude=new_point['longitude'], depth=start_point.depth)

        # We don't want to reverse further than the current spatial buffer, because we will reindex the
        # source file everytime we reverse, which will slow down the calculations considerably.
        while (not self._spatial_query_object.contains(new_loc.point) or self.intersect(start_point=nudge_loc.point, end_point=new_loc.point)) and count < 6:
            changing_distance /= 2
            new_point = AsaGreatCircle.great_circle(distance=changing_distance, azimuth=random_azimuth, start_point=hit_point)
            new_loc = Location4D(latitude=new_point['latitude'], longitude=new_point['longitude'], depth=start_point.depth)
            count += 1

        # We tried 10 times and the particle was still on shore, return the point the particle started from.
        # No randomization.
        if count == 6:
            logger.warn("LOOK: Could not react particle with shoreline.  Assuming particle did not move from original location")
            return start_point

        #logger.info("Reaction time: %f" % (time.clock() - st))
        return new_loc


class ShorelineFile(Shoreline):
    """
    Shoreline backed by a shapefile on disk.
    """
    def __init__(self, file=None, path=None, **kwargs):
        """
            Optional named arguments:
            * file (local path to OGC complient file)
            * path (used instead of file)

            MUST BE land polygons!!
        """
        if path is not None:
            self._file = os.path.normpath(path)
        elif file is not None:
            self._file = os.path.normpath(file)
        else:
            self._file = os.path.normpath(os.path.join(__file__, "../../resources/shoreline/global/10m_land.shp"))

        self._source = fiona.open(self._file)

        super(ShorelineFile, self).__init__(**kwargs)

    def close(self):
        super(ShorelineFile, self).close()
        self._source.close()

    def get_capabilities(self):
        """
        Gets capabilities.

        This is a simulation of a GetCapabilities WFS request. Returns a python dict
        with LatLongBoundingBox and Name keys defined.
        """
        d = {}
        d['LatLongBoundingBox'] = box(*self._source.bounds)
        d['Name']               = self._file.split('/')[-1].split('.')[0]

        return d

    def get_feature_type_info(self):
        """
        Gets FeatureType as a python dict.

        On ShorelineFile this is a passthrough to the simulated get_capabilities.
        """
        return self.get_capabilities()

    def get_geoms_for_bounds(self, bounds):
        """
        Helper method to get geometries within a certain bounds (as WKT).

        Returns GeoJSON (loaded as a list of python dictionaries).
        """
        return [asShape(e[1]['geometry']) for e in self._source.items(bbox=bounds)]

    def index(self, point=None, spatialbuffer=None):
        """
            This queries the shapefile around a buffer of a point
            The results of this spatial query are used for shoreline detection.

            Using the entire shapefile without the spatial query takes over
            30 times the time with world land polygons.

        """
        spatialbuffer = spatialbuffer or self._spatialbuffer
        self._spatial_query_object = None
        geoms                      = []

        if point:
            self._spatial_query_object = point.buffer(spatialbuffer)
            geoms = self.get_geoms_for_bounds(self._spatial_query_object.envelope.bounds)

        self._geoms = []

        # The _geoms should be only Polygons, not MultiPolygons
        for geom in geoms:
            try:
                if isinstance(geom, Polygon):
                    self._geoms.append(geom)
                elif isinstance(geom, MultiPolygon):
                    for poly in geom:
                        self._geoms.append(poly)
            except:
                logger.warn("Could not find valid geometry in shoreline element.  Point: %s, Buffer: %s" % (str(point), str(spatialbuffer)))


class ShorelineWFS(Shoreline):
    """
    Shoreline backed by a WFS server.
    """

    def __init__(self, path=None, wfs_server=None, feature_name=None, **kwargs):
        self._wfs_server   = path or wfs_server
        self._feature_name = feature_name

        super(ShorelineWFS, self).__init__(**kwargs)

    def get_capabilities(self):
        """
        Gets capabilities.

        Queries WFS server for its capabilities. Internally used by get_feature_type_info.
        """
        params = {'service'      : 'WFS',
                  'request'      : 'GetCapabilities',
                  'version'      : '1.0.0'}

        caps_response = requests.get(self._wfs_server, params=params)
        caps_response.raise_for_status()

        return ET.fromstring(caps_response.content)

    def get_feature_type_info(self):
        """
        Gets FeatureType as a python dict.

        Transforms feature_name info into python dict.
        """
        caps = self.get_capabilities()
        if caps is None:
            return None

        el = caps.find('{http://www.opengis.net/wfs}FeatureTypeList')
        for e in el.findall('{http://www.opengis.net/wfs}FeatureType'):
            if e.find('{http://www.opengis.net/wfs}Name').text == self._feature_name:
                # transform into python dict
                # <Name>sample</Name>
                # <Abstract/>
                # <LatLongBoundingBox maxx="1" maxy="5" ... />
                #
                # becomes:
                #
                # {'Name'               :'sample',
                #  'Abtract'            : None,
                #  'LatLongBoundingBox' : {'maxx':1, 'maxy':5 ... }}
                #
                d = {sube.tag[28:] : sube.text or sube.attrib or None for sube in e.getchildren()}

                # transform LatLongBoundingBox into a Shapely box
                llbb = {k : round(float(v), 4) for k, v in d['LatLongBoundingBox'].items()}
                d['LatLongBoundingBox'] = box(llbb['minx'], llbb['miny'], llbb['maxx'], llbb['maxy'])
                return d

        return None

    def get_geoms_for_bounds(self, bounds):
        """
        Helper method to get geometries within a certain bounds (as WKT).

        Returns GeoJSON (loaded as a list of python dictionaries).
        """

        params = {'service'      : 'WFS',
                  'request'      : 'GetFeature',
                  'typeName'     : self._feature_name,
                  'outputFormat' : 'json',
                  'version'      : '1.0.0',
                  'bbox'         : ','.join((str(b) for b in wkt.loads(bounds).bounds))}

        raw_geojson_response = requests.get(self._wfs_server, params=params)
        raw_geojson_response.raise_for_status()
        gj = raw_geojson_response.json()

        return [asShape(g['geometry']) for g in gj['features']]

    def index(self, point=None, spatialbuffer=None):
        spatialbuffer              = spatialbuffer or self._spatialbuffer
        self._spatial_query_object = None
        geoms                      = []

        if point:
            self._spatial_query_object = point.buffer(spatialbuffer)
            bounds                     = point.buffer(spatialbuffer).envelope.wkt
            geoms                      = self.get_geoms_for_bounds(bounds)

        self._geoms = []

        for geom in geoms:
            try:
                if isinstance(geom, Polygon):
                    self._geoms.append(geom)
                elif isinstance(geom, MultiPolygon):
                    for poly in geom:
                        self._geoms.append(poly)
            except:
                logger.warn("Could not find valid geometry in shoreline element.  Point: %s, Buffer: %s" % (str(point), str(spatialbuffer)))



import sys
import json
import redis
import logging
import multiprocessing
import traceback
from datetime import datetime
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from paegan.transport.forcers import BaseForcer
from paegan.transport.controllers import BaseModelController, Runner
from paegan.logger import logger
import paegan.transport.export as ex


def particle_runner(part, model):

    from paegan.logger import logger
    logger.setLevel(logging.PROGRESS)

    from paegan.logger.redis_handler import RedisHandler
    rhandler = RedisHandler(model.redis_log_channel, model.redis_url)
    rhandler.setLevel(logging.PROGRESS)
    logger.addHandler(rhandler)

    try:
        redis_connection = redis.from_url(model.redis_url)
        forcer = BaseForcer(model.hydrodataset,
                            particle=part,
                            common_variables=model.common_variables,
                            times=model.times,
                            start_time=model.start,
                            models=model._models,
                            release_location_centroid=model.reference_location.point,
                            usebathy=model._use_bathymetry,
                            useshore=model._use_shoreline,
                            usesurface=model._use_seasurface,
                            reverse_distance=model.reverse_distance,
                            bathy_path=model.bathy_path,
                            shoreline_path=model.shoreline_path,
                            shoreline_feature=model.shoreline_feature,
                            time_method=model.time_method,
                            redis_url=model.redis_url,
                            redis_results_channel=model.redis_results_channel,
                            shoreline_index_buffer=model.shoreline_index_buffer)
        forcer.run()
    except Exception:
        logger.exception(traceback.format_exc())
        redis_connection.publish(model.redis_results_channel, json.dumps({"status" : "FAILED", "uid" : part.uid }))
    else:
        redis_connection.publish(model.redis_results_channel, json.dumps({"status" : "COMPLETED", "uid" : part.uid }))


class DistributedModelController(BaseModelController):

    def __init__(self, **kwargs):
        super(DistributedModelController, self).__init__(**kwargs)
        self.thread_result_listener = True

    def start_tasks(self, **kwargs):
        logger.progress((5, 'Running model'))
        rc = redis.from_url(self.redis_url)
        if kwargs.get('task_queue_call'):
            for p in self.particles:
                try:
                    kwargs.get('task_queue_call')(func=particle_runner, args=(p, self,))
                except Exception:
                    logger.exception(traceback.format_exc())
                    rc.publish(self.redis_results_channel, json.dumps({"status" : "FAILED", "uid" : p.uid }))
            return True
        else:
            tasks = []
            for p in self.particles:
                f = BaseForcer(self.hydrodataset,
                               particle=p,
                               common_variables=self.common_variables,
                               times=self.times,
                               start_time=self.start,
                               models=self._models,
                               release_location_centroid=self.reference_location.point,
                               usebathy=self._use_bathymetry,
                               useshore=self._use_shoreline,
                               usesurface=self._use_seasurface,
                               reverse_distance=self.reverse_distance,
                               bathy_path=self.bathy_path,
                               shoreline_path=self.shoreline_path,
                               shoreline_feature=self.shoreline_feature,
                               time_method=self.time_method,
                               redis_url=self.redis_url,
                               redis_results_channel=self.redis_results_channel,
                               shoreline_index_buffer=self.shoreline_index_buffer)
                tasks.append(f)
            self.pool = multiprocessing.Pool()
            return self.pool.imap(Runner(), tasks)

        return False

    def setup_run(self, hydrodataset, **kwargs):
        from paegan.logger.redis_handler import RedisHandler
        self.redis_url             = kwargs.get("redis_url")
        self.redis_log_channel     = kwargs.get("redis_log_channel")
        self.redis_results_channel = kwargs.get("redis_results_channel")
        rhandler = RedisHandler(self.redis_log_channel, self.redis_url)
        rhandler.setLevel(logging.PROGRESS)
        logger.addHandler(rhandler)

        super(DistributedModelController, self).setup_run(hydrodataset, **kwargs)

    def listen_for_results(self, output_h5_file, total_particles):
            # Create output file (hdf5)
            particles_finished = 0
            results = ex.ResultsPyTable(output_h5_file)

            res = urlparse(self.redis_url)
            redis_pool = redis.ConnectionPool(host=res.hostname, port=res.port, db=res.path[1:])
            r = redis.Redis(connection_pool=redis_pool)

            pubsub = r.pubsub()
            pubsub.subscribe(self.redis_results_channel)
            for msg in pubsub.listen():

                if msg['type'] != "message":
                    continue

                if msg["data"] == "FINISHED":
                    break

                try:
                    json_msg = json.loads(msg["data"])
                    if json_msg.get("status", None):
                        #  "COMPLETED" or "FAILED" when a particle finishes
                        particles_finished += 1
                        percent_complete = 90. * (float(particles_finished) / float(total_particles)) + 5  # Add the 5 progress that was used prior to the particles starting (controller)
                        r.publish(self.redis_log_channel, json.dumps({"time" : datetime.utcnow().isoformat(), "level" : "progress", "value" : percent_complete, "message" : "Particle #%s %s!" % (particles_finished, json_msg.get("status"))}))
                        if particles_finished == total_particles:
                            break
                    else:
                        # Write to HDF file
                        results.write(json_msg)
                except Exception:
                    pass

            pubsub.close()
            results.compute()
            results.close()
            redis_pool.disconnect()
            sys.exit()



import json
from random import gauss, uniform

class Capability(object):

    def __init__(self, **kwargs):

        if 'json' in kwargs or 'data' in kwargs:
            data = {}
            try:
                data = json.loads(kwargs['json'])
            except Exception:
                try:
                    data = kwargs.get('data')
                except Exception:
                    pass

            self.vss = data.get('vss', None)
            self.swim_turning = data.get('swim_turning', None)
            self.non_swim_turning = data.get('nonswim_turning', None)
            self.variance = data.get('variance', 0)
            # We initialize the calculated VSS here.  This is not
            # Recalculated on access.  It needs to be manually recalculated
            # for each Capability.
            self.calculated_vss = self.calculate_vss(method=kwargs.get('method', None))

    def calculate_vss(self, method=None):
        """
        Calculate the vertical swimming speed of this behavior.
        Takes into account the vertical swimming speed and the
        variance.

        Parameters:
            method:  "gaussian" (default) or "random"
                     "random" (vss - variance) < X < (vss + variance)
        """
        if self.variance == float(0):
            return self.vss
        else:
            # Calculate gausian distribution and return
            if method == "gaussian" or method is None:
                return gauss(self.vss, self.variance)
            elif method == "random":
                return uniform(self.vss - self.variance, self.vss + self.variance)
            else:
                raise ValueError("Method of vss calculation not recognized, please use 'gaussian' or 'random'")
    def get_calculated_vss(self):
        return self._calculated_vss
    def set_calculated_vss(self, cvss):
        self._calculated_vss = float(cvss)
    calculated_vss = property(get_calculated_vss, set_calculated_vss)

    def get_vss(self):
        return self._vss
    def set_vss(self, vss):
        self._vss = float(vss)
    vss = property(get_vss, set_vss)

    def get_variance(self):
        return self._variance
    def set_variance(self, variance):
        self._variance = float(variance)
    variance = property(get_variance, set_variance)

    def get_nonswim_turning(self):
        return self._non_swim_turning
    def set_nonswim_turning(self, turning):
        self._non_swim_turning = turning
    non_swim_turning = property(get_nonswim_turning, set_nonswim_turning)

    def get_swim_turning(self):
        return self._swim_turning
    def set_swim_turning(self, turning):
        self._swim_turning = turning
    swim_turning = property(get_swim_turning, set_swim_turning)

    def __str__(self):
        return """
            vss (m/s): %s (+/- %s)
            calculated_vss (m/s): %s
            swim_turning: %s
            non_swim_turning: %s
        """ % (self.vss, self.variance, self.calculated_vss, self.swim_turning, self.non_swim_turning)



import os
import glob
import math
import zipfile
from shapely.geometry import MultiPoint, Point, mapping
from datetime import datetime

# NetCDF
import netCDF4
import pytz

from fiona import collection

from collections import OrderedDict

from paegan.logger import logger

import geojson

import tables
from tables import *


# Pytables representation of a model run
class ModelResultsTable(IsDescription):
    particle    = UInt8Col()
    time        = Time32Col()
    latitude    = Float32Col()
    longitude   = Float32Col()
    depth       = Float32Col()
    u_vector    = Float32Col()
    v_vector    = Float32Col()
    w_vector    = Float32Col()
    temperature = Float32Col()
    salinity    = Float32Col()
    age         = Float32Col()
    lifestage   = UInt8Col()
    progress    = Float32Col()
    settled     = BoolCol()
    halted      = BoolCol()
    dead        = BoolCol()


class ResultsPyTable(object):
    def __init__(self, output_file):
        self._file  = open_file(output_file, mode="w", title="Model run output")
        self._root  = self._file.create_group("/", "trajectories", "Trajectory Data")
        self._table = self._file.create_table(self._root, "model_results", ModelResultsTable, "Model Results")
        self._table.autoindex = False
        self._table.cols.particle.create_index()
        self._table.cols.time.create_index()
        self._table.cols.latitude.create_index()
        self._table.cols.longitude.create_index()

    def write(self, data):
        record = self._table.row
        for k, v in data.items():
            try:
                record[k] = v
            except Exception:
                # No column named "k", so don't add the data
                pass

        record.append()

    def trackline(self):
        pass

    def metadata(self):
        pass

    def compute(self):
        self.trackline()
        self.metadata()

    def close(self):
        self._table.flush()
        self._table.reindex()
        self._file.close()


class Export(object):
    @classmethod
    def export(cls, **kwargs):
        raise("Please implement the export method of your Export class.")


class H5TracklineWithPoints(Export):
    @classmethod
    def export(cls, folder, h5_file):
        with tables.open_file(h5_file, mode="r") as h5:
            table = h5.root.trajectories.model_results
            timestamps = sorted(list(set([ x["time"] for x in table.iterrows() ])))

            pts = []
            features = []
            for i, ts in enumerate(timestamps):
                points = MultiPoint([ Point(x['longitude'], x['latitude']) for x in table.where("""time == %s""" % ts) if x["latitude"] and x["longitude"] ])
                cp = points.centroid.coords[0]
                geo_pt = geojson.Point(cp)
                pts.append(cp)
                feat = geojson.Feature(id=i, geometry=geo_pt, properties={ "time" : datetime.utcfromtimestamp(ts).replace(tzinfo=pytz.utc).isoformat() })
                features.append(feat)

            geo_ls = geojson.LineString(pts)
            features.append(geojson.Feature(geometry=geo_ls, id='path'))

            fc = geojson.FeatureCollection(features)

            if not os.path.exists(folder):
                os.makedirs(folder)
            filepath = os.path.join(folder, "full_trackline.geojson")
            with open(filepath, "wb") as r:
                r.write(geojson.dumps(fc).encode('utf-8'))


class H5Trackline(Export):
    @classmethod
    def export(cls, folder, h5_file):
        with tables.open_file(h5_file, mode="r") as h5:
            table = h5.root.trajectories.model_results
            timestamps = sorted(list(set([ x["time"] for x in table.iterrows() ])))

            pts = []
            for i, ts in enumerate(timestamps):
                points = MultiPoint([ Point(x['longitude'], x['latitude']) for x in table.where("""time == %s""" % ts) if x["latitude"] and x["longitude"] ])
                pts.append(points.centroid.coords[0])
            geo_ls = geojson.LineString(pts)
            feat = geojson.Feature(geometry=geo_ls, id='path')

            if not os.path.exists(folder):
                os.makedirs(folder)
            filepath = os.path.join(folder, "simple_trackline.geojson")
            with open(filepath, "wb") as r:
                r.write(geojson.dumps(feat).encode('utf-8'))


class H5ParticleTracklines(Export):
    @classmethod
    def export(cls, folder, h5_file):
        with tables.open_file(h5_file, mode="r") as h5:
            table = h5.root.trajectories.model_results
            particles = sorted(list(set([ x["particle"] for x in table.iterrows() ])))

            features = []
            for puid in particles:
                points = [ (x["time"], (x['longitude'], x['latitude'])) for x in table.where("""particle == %s""" % puid) if x["latitude"] and x["longitude"] ]

                geo_ls = geojson.LineString( [ x[1] for x in points ] )
                times  = [ datetime.utcfromtimestamp(x[0]).replace(tzinfo=pytz.utc).isoformat() for x in points ]

                feat = geojson.Feature(geometry=geo_ls, id=puid, properties={ "particle" : puid, "times" : times })
                features.append(feat)

            fc = geojson.FeatureCollection(features)

            if not os.path.exists(folder):
                os.makedirs(folder)
            filepath = os.path.join(folder, "particle_tracklines.geojson")
            with open(filepath, "wb") as r:
                r.write(geojson.dumps(fc).encode('utf-8'))


class H5ParticleMultiPoint(Export):
    @classmethod
    def export(cls, folder, h5_file):
        with tables.open_file(h5_file, mode="r") as h5:
            table = h5.root.trajectories.model_results
            particles = sorted(list(set([ x["particle"] for x in table.iterrows() ])))

            features = []
            for puid in particles:
                points = [ (x["time"], (x['longitude'], x['latitude'])) for x in table.where("""particle == %s""" % puid) if x["latitude"] and x["longitude"] ]

                geo_mp = geojson.MultiPoint( [ x[1] for x in points ] )
                times  = [ x[0] for x in points ]

                feat = geojson.Feature(geometry=geo_mp, id=puid, properties={ "particle" : puid, "time" : times })
                features.append(feat)

            fc = geojson.FeatureCollection(features)

            if not os.path.exists(folder):
                os.makedirs(folder)
            filepath = os.path.join(folder, "particle_multipoint.geojson")
            with open(filepath, "wb") as r:
                r.write(geojson.dumps(fc).encode('utf-8'))


class H5GDALShapefile(Export):
    @classmethod
    def export(cls, folder, h5_file):
        shape_schema = {'geometry': 'Point',
                        'properties': OrderedDict([('particle',     'int'),
                                                   ('date',         'str'),
                                                   ('latitude',     'float'),
                                                   ('longitude',    'float'),
                                                   ('depth',        'float'),
                                                   ('u_vector',     'float'),
                                                   ('v_vector',     'float'),
                                                   ('w_vector',     'float'),
                                                   ('temp',         'float'),
                                                   ('salinity',     'float'),
                                                   ('age',          'float'),
                                                   ('settled',      'str'),
                                                   ('dead',         'str'),
                                                   ('halted',       'str'),
                                                   ('lifestage',    'int'),
                                                   ('progress',     'float')])}
        shape_crs = {'no_defs': True, 'ellps': 'WGS84', 'datum': 'WGS84', 'proj': 'longlat'}

        if not os.path.exists(folder):
            os.makedirs(folder)
        filepath = os.path.join(folder, "particles.shp")

        with tables.open_file(h5_file, mode="r") as h5:
            table = h5.root.trajectories.model_results

            with collection(filepath, "w", driver='ESRI Shapefile', schema=shape_schema, crs=shape_crs) as shape:
                for r in table.iterrows():
                    shape.write({'geometry': mapping(Point(r["longitude"], r["latitude"])),
                                 'properties': OrderedDict([('particle', r["particle"]),
                                                            ('date', str(datetime.utcfromtimestamp(r["time"]).isoformat())),
                                                            ('latitude', float(r["latitude"])),
                                                            ('longitude', float(r["longitude"])),
                                                            ('depth', float(r["depth"])),
                                                            ('temp', float(r["temperature"])),
                                                            ('salinity', float(r["salinity"])),
                                                            ('u_vector', float(r["u_vector"])),
                                                            ('v_vector', float(r["v_vector"])),
                                                            ('w_vector', float(r["w_vector"])),
                                                            ('settled', str(r["settled"])),
                                                            ('dead', str(r["dead"])),
                                                            ('halted', str(r["halted"])),
                                                            ('age', float(r["age"])),
                                                            ('lifestage' , int(r["lifestage"])),
                                                            ('progress' , float(r["progress"]))])})

        # Zip the output
        shpzip = zipfile.ZipFile(os.path.join(folder, "h5shape.shp.zip"), mode='w')
        for f in glob.glob(os.path.join(folder, "particles*")):
            shpzip.write(f, os.path.basename(f))
            os.remove(f)
        shpzip.close()


class NetCDF(Export):
    @classmethod
    def export(cls, folder, particles, datetimes, summary, **kwargs):
        """
            Export particle data to CF trajectory convention
            netcdf file
        """
        time_units = 'seconds since 1990-01-01 00:00:00'

        # Create netcdf file, overwrite existing
        if not os.path.exists(folder):
            os.makedirs(folder)

        filepath = os.path.join(folder, 'trajectories.nc')
        nc = netCDF4.Dataset(filepath, 'w')
        # Create netcdf dimensions
        nc.createDimension('time', None)
        nc.createDimension('particle', None)

        fillvalue = -9999.9

        # Create netcdf variables
        time = nc.createVariable('time', 'i', ('time',))
        part = nc.createVariable('particle', 'i', ('particle',))
        depth = nc.createVariable('depth', 'f', ('time', 'particle'))
        lat = nc.createVariable('lat', 'f', ('time', 'particle'), fill_value=fillvalue)
        lon = nc.createVariable('lon', 'f', ('time', 'particle'), fill_value=fillvalue)
        salt = nc.createVariable('salt', 'f', ('time', 'particle'), fill_value=fillvalue)
        temp = nc.createVariable('temp', 'f', ('time', 'particle'), fill_value=fillvalue)
        u = nc.createVariable('u', 'f', ('time', 'particle'), fill_value=fillvalue)
        v = nc.createVariable('v', 'f', ('time', 'particle'), fill_value=fillvalue)
        w = nc.createVariable('w', 'f', ('time', 'particle'), fill_value=fillvalue)
        settled = nc.createVariable('settled', 'f', ('time', 'particle'), fill_value=fillvalue)
        dead = nc.createVariable('dead', 'f', ('time', 'particle'), fill_value=fillvalue)
        halted = nc.createVariable('halted', 'f', ('time', 'particle'), fill_value=fillvalue)

        # Loop through locations in each particle,
        # add to netcdf file
        for j, particle in enumerate(particles):
            part[j] = particle.uid
            i = 0

            normalized_locations = particle.normalized_locations(datetimes)
            normalized_temps = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.temps]
            normalized_salts = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.salts]
            normalized_u = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.u_vectors]
            normalized_v = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.v_vectors]
            normalized_w = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.w_vectors]
            normalized_settled = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.settles]
            normalized_dead = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.deads]
            normalized_halted = [x if x is not None and not math.isnan(x) else fillvalue for x in particle.halts]

            if len(normalized_locations) != len(normalized_temps):
                logger.info("No temperature being added to netcdf.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_temps = [fillvalue] * len(normalized_locations)

            if len(normalized_locations) != len(normalized_salts):
                logger.info("No salinity being added to netcdf.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_salts = [fillvalue] * len(normalized_locations)

            if len(normalized_locations) != len(normalized_u):
                logger.info("No U being added to netcdf.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_u = [fillvalue] * len(normalized_locations)

            if len(normalized_locations) != len(normalized_v):
                logger.info("No V being added to netcdf.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_v = [fillvalue] * len(normalized_locations)

            if len(normalized_locations) != len(normalized_w):
                logger.info("No W being added to netcdf.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_w = [fillvalue] * len(normalized_locations)

            if len(normalized_locations) != len(normalized_settled):
                logger.info("No Settled being added to shapefile.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_settled = [fillvalue] * len(normalized_locations)

            if len(normalized_locations) != len(normalized_dead):
                logger.info("No Dead being added to shapefile.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_dead = [fillvalue] * len(normalized_locations)

            if len(normalized_locations) != len(normalized_halted):
                logger.info("No Halted being added to shapefile.")
                # Create list of 'fillvalue' equal to the length of locations
                normalized_halted = [fillvalue] * len(normalized_locations)

            for loc, _temp, _salt, _u, _v, _w, _settled, _dead, _halted in zip(normalized_locations, normalized_temps, normalized_salts, normalized_u, normalized_v, normalized_w, normalized_settled, normalized_dead, normalized_halted):

                if j == 0:
                    time[i] = int(round(netCDF4.date2num(loc.time, time_units)))
                depth[i, j] = loc.depth
                lat[i, j] = loc.latitude
                lon[i, j] = loc.longitude
                salt[i, j] = _salt
                temp[i, j] = _temp
                u[i, j] = _u
                v[i, j] = _v
                w[i, j] = _w
                settled[i, j] = _settled
                dead[i, j] = _dead
                halted[i, j] = _halted
                i += 1

        # Variable attributes
        depth.coordinates = "time particle lat lon"
        depth.standard_name = "depth_below_sea_surface"
        depth.units = "m"
        depth.POSITIVE = "up"
        depth.positive = "up"

        salt.coordinates = "time particle lat lon"
        salt.standard_name = "sea_water_salinity"
        salt.units = "psu"

        temp.coordinates = "time particle lat lon"
        temp.standard_name = "sea_water_temperature"
        temp.units = "degrees_C"

        u.coordinates = "time particle lat lon"
        u.standard_name = "eastward_sea_water_velocity"
        u.units = "m/s"

        v.coordinates = "time particle lat lon"
        v.standard_name = "northward_sea_water_velocity"
        v.units = "m/s"

        w.coordinates = "time particle lat lon"
        w.standard_name = "upward_sea_water_velocity"
        w.units = "m/s"

        settled.coordinates = "time particle lat lon"
        settled.description = "Is the particle settled"
        settled.standard_name = "particle_settled"

        dead.coordinates = "time particle lat lon"
        dead.description = "Is the particle dead"
        dead.standard_name = "particle_dead"

        halted.coordinates = "time particle lat lon"
        halted.description = "Is the particle prevented from being forced by currents"
        halted.standard_name = "particle_halted"

        time.units = time_units
        time.standard_name = "time"

        lat.units = "degrees_north"

        lon.units = "degrees_east"

        part.cf_role = "trajectory_id"

        # Global attributes
        nc.featureType = "trajectory"
        nc.summary = str(summary)
        for key in kwargs:
            nc.__setattr__(key, kwargs.get(key))

        nc.sync()
        nc.close()
